from PySide6.QtGui import Qt
from PySide6.QtWidgets import (
    QWidget,
    QHBoxLayout,
    QComboBox,
    QStackedWidget,
    QLabel,
    QSizePolicy,
)

from caqtus.device.output_transform import EvaluableOutput
from caqtus.types.expression import Expression
from ._expression_editor import ExpressionEditor
from ._value_editor import ValueEditor


class OutputTransformEditor(ValueEditor[EvaluableOutput]):
    def __init__(self) -> None:
        self._widget = QWidget()
        layout = QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        self._widget.setLayout(layout)
        self._combo_box = QComboBox()
        self._combo_box.addItem("Expression")
        self._combo_box.addItem("Tree")
        layout.addWidget(self._combo_box)
        layout.setAlignment(self._combo_box, Qt.AlignmentFlag.AlignTop)

        self._expression_editor = ExpressionEditor()

        self._tree = QLabel("Not implemented")
        self._stacked_widget = QStackedWidget()
        self._stacked_widget.addWidget(self._expression_editor.widget)
        self._stacked_widget.addWidget(self._tree)
        self._combo_box.currentIndexChanged.connect(
            self._stacked_widget.setCurrentIndex
        )

        layout.addWidget(self._stacked_widget)
        self._stacked_widget.setSizePolicy(
            QSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Fixed)
        )
        self._combo_box.setCurrentIndex(0)

    def set_value(self, value: EvaluableOutput) -> None:
        if isinstance(value, Expression):
            self._combo_box.setCurrentIndex(0)
            self._expression_editor.set_value(value)
        else:
            self._combo_box.setCurrentIndex(1)

    @property
    def widget(self) -> QWidget:
        return self._widget

    def set_editable(self, editable: bool) -> None:
        self._combo_box.setEnabled(editable)

    def read_value(self) -> EvaluableOutput:
        if self._combo_box.currentIndex() == 0:
            return self._expression_editor.read_value()
        else:
            raise NotImplementedError("Tree output transform not implemented")



"""Utilities to work with base units.

Base units are units that are expressed only as power of SI units.
For example, `kg`, `m/s`, 'kg*m^2/s**2` are base units, but `mg`, `km/h`, 'J' or `dB`
are not.

Non-base units are useful for human readability, but some operations are ill-defined
when using them, especially for non-linear operations.
It is strongly recommended to convert all values to base units as soon as possible to
avoid ambiguities.
"""

from typing import overload, Optional

from typing_extensions import TypeIs

from ._units import (
    Unit,
    BaseUnit,
    Quantity,
    dimensionless,
    Magnitude,
)


def is_in_base_units(units: Unit) -> TypeIs[BaseUnit]:
    """Check if the units is only expressed in terms of base SI units.

    For example, `kg`, `m/s` are expressed in terms of base units, but `mg`, `km/h` or
    `dB` are not.

    Args:
        units: The units to check.

    Returns:
        True if the units are expressed in the base units of the registry.
    """

    return units.to_base() == units


def is_base_quantity[M: Magnitude](value: Quantity[M]) -> TypeIs[Quantity[M, BaseUnit]]:
    """Check if the given value is expressed in base units."""

    return is_in_base_units(value.units)


@overload
def convert_to_base_units[M: Magnitude](magnitude: M, unit: None) -> tuple[M, None]: ...


@overload
def convert_to_base_units[
    M: Magnitude
](magnitude: M, unit: Unit) -> tuple[M, Optional[BaseUnit]]: ...


def convert_to_base_units(
    magnitude: Magnitude, unit: Optional[Unit]
) -> tuple[Magnitude, Optional[BaseUnit]]:
    """Convert values into base units.

    Args:
        magnitude: The value to convert.
            Can be a scalar or an array of values.
        unit: The unit in which the magnitude is expressed.

    Returns:
        The magnitude in base units and the base units.
    """

    if unit is None:
        return magnitude, None
    else:
        quantity = Quantity(magnitude, unit)
        in_base_units = quantity.to_base_units()
        magnitude_in_base_units = in_base_units.magnitude
        base_units = in_base_units.units
        if base_units == dimensionless:
            base_units = None
        else:
            assert is_in_base_units(base_units)
        return magnitude_in_base_units, base_units



from __future__ import annotations

import multiprocessing.managers
import time
from collections.abc import Mapping
from typing import Optional

from caqtus.device import DeviceName, DeviceConfiguration
from caqtus.session import ExperimentSessionMaker, PureSequencePath
from caqtus.types.parameter import ParameterNamespace
from .manager import ExperimentManager, Procedure, LocalExperimentManager
from ..device_manager_extension import DeviceManagerExtensionProtocol
from ..sequence_execution import ShotRetryConfig

experiment_manager: Optional[LocalExperimentManager] = None


class ExperimentManagerProxy(ExperimentManager, multiprocessing.managers.BaseProxy):
    _exposed_ = ("create_procedure", "interrupt_running_procedure")
    _method_to_typeid_ = {
        "create_procedure": "ProcedureProxy",
    }

    def create_procedure(
        self, procedure_name: str, acquisition_timeout: Optional[float] = None
    ) -> ProcedureProxy:
        return self._callmethod("create_procedure", (procedure_name,))  # type: ignore

    def interrupt_running_procedure(self) -> bool:
        return self._callmethod("interrupt_running_procedure", ())  # type: ignore

    def __repr__(self):
        return f"<ExperimentManagerProxy at {hex(id(self))}>"


class ProcedureProxy(Procedure, multiprocessing.managers.BaseProxy):
    """Proxy for a procedure running in a different process.

    This object behaves like a :class:`Procedure` object and should be used like it, but
    it forwards all method calls to an actual procedure object running in a different
    process.
    """

    _exposed_ = (
        "__enter__",
        "__exit__",
        "is_active",
        "is_running_sequence",
        "sequences",
        "exception",
        "start_sequence",
        "interrupt_sequence",
        "run_sequence",
        "__str__",
    )
    _method_to_typeid_ = {"__enter__": "ProcedureProxy"}

    def __enter__(self):
        return self._callmethod("__enter__", ())

    def __exit__(self, exc_type, exc_val, exc_tb):
        return self._callmethod("__exit__", (None, exc_val, None))

    def __repr__(self):
        return f"<{self.__class__.__name__}('{self}') at {hex(id(self))}>"

    def __str__(self):
        result = self._callmethod("__str__", ())
        assert isinstance(result, str)
        return result

    def is_active(self) -> bool:
        return self._callmethod("is_active", ())  # type: ignore

    def is_running_sequence(self) -> bool:
        return self._callmethod("is_running_sequence", ())  # type: ignore

    def sequences(self) -> list[PureSequencePath]:
        return self._callmethod("sequences", ())  # type: ignore

    def exception(self) -> Optional[Exception]:
        while self.is_running_sequence():
            time.sleep(10e-3)
        return self._callmethod("exception", ())  # type: ignore

    def start_sequence(
        self,
        sequence: PureSequencePath,
        global_parameters: Optional[ParameterNamespace] = None,
        device_configurations: Optional[
            Mapping[DeviceName, DeviceConfiguration]
        ] = None,
    ) -> None:
        return self._callmethod(
            "start_sequence",
            (sequence, global_parameters, device_configurations),
        )

    def interrupt_sequence(self) -> bool:
        return self._callmethod("interrupt_sequence", ())  # type: ignore

    def run_sequence(
        self,
        sequence: PureSequencePath,
        global_parameters: Optional[ParameterNamespace] = None,
        device_configurations: Optional[
            Mapping[DeviceName, DeviceConfiguration]
        ] = None,
    ) -> None:
        # Here we can't just call the remote method `run_sequence`.
        # This is because this method takes a long time to return, since it waits for
        # the sequence to finish.
        # This means that we can't even raise KeyboardInterrupt while waiting for the
        # method to return.
        self.start_sequence(sequence, global_parameters, device_configurations)
        while self.is_running_sequence():
            time.sleep(10e-3)
        if exception := self.exception():
            raise exception


class _MultiprocessingServerManager(multiprocessing.managers.BaseManager):
    pass


def _get_experiment_manager() -> LocalExperimentManager:
    if experiment_manager is None:
        raise RuntimeError("Experiment manager not initialized")
    return experiment_manager


def _enter_experiment_manager() -> None:
    assert experiment_manager is not None
    experiment_manager.__enter__()


def _exit_experiment_manager(exc_value) -> None:
    assert experiment_manager is not None
    experiment_manager.__exit__(type(exc_value), exc_value, exc_value.__traceback__)


def _create_experiment_manager(
    session_maker: ExperimentSessionMaker,
    device_manager_extension: DeviceManagerExtensionProtocol,
    shot_retry_config: Optional[ShotRetryConfig] = None,
) -> None:
    global experiment_manager
    experiment_manager = LocalExperimentManager(
        session_maker=session_maker,
        shot_retry_config=shot_retry_config,
        device_manager_extension=device_manager_extension,
    )


_MultiprocessingServerManager.register(
    "create_experiment_manager", _create_experiment_manager, ExperimentManagerProxy
)
_MultiprocessingServerManager.register(
    "get_experiment_manager", _get_experiment_manager, ExperimentManagerProxy
)
_MultiprocessingServerManager.register(
    "enter_experiment_manager", _enter_experiment_manager
)
_MultiprocessingServerManager.register(
    "exit_experiment_manager", _exit_experiment_manager
)
_MultiprocessingServerManager.register("ProcedureProxy", None, ProcedureProxy)


class RemoteExperimentManagerServer:
    session_maker: Optional[ExperimentSessionMaker] = None

    def __init__(
        self,
        address: tuple[str, int],
        authkey: bytes,
        session_maker: ExperimentSessionMaker,
        device_manager_extension: DeviceManagerExtensionProtocol,
        shot_retry_config: Optional[ShotRetryConfig] = None,
    ):
        self._session_maker = session_maker
        self._multiprocessing_manager = _MultiprocessingServerManager(
            address=address, authkey=authkey
        )
        self._shot_retry_config = shot_retry_config
        self._shot_retry_config = shot_retry_config
        self._device_manager_extension = device_manager_extension

    def __enter__(self):
        self._multiprocessing_manager.start()
        self._multiprocessing_manager.create_experiment_manager(  # type: ignore
            self._session_maker,
            self._device_manager_extension,
            self._shot_retry_config,
        )
        self._multiprocessing_manager.enter_experiment_manager()  # type: ignore
        return self

    @staticmethod
    def serve_forever():
        while True:
            time.sleep(100e-3)

    def __exit__(self, exc_type, exc_value, traceback):
        self._multiprocessing_manager.exit_experiment_manager(exc_value)  # type: ignore
        return self._multiprocessing_manager.__exit__(exc_type, exc_value, traceback)


class _MultiprocessingClientManager(multiprocessing.managers.BaseManager):
    pass


_MultiprocessingClientManager.register(
    "get_experiment_manager", None, ExperimentManagerProxy
)

_MultiprocessingClientManager.register("ProcedureProxy", None, ProcedureProxy)


class RemoteExperimentManagerClient:
    def __init__(self, address: tuple[str, int], authkey: bytes):
        self._multiprocessing_manager = _MultiprocessingClientManager(
            address=address, authkey=authkey
        )
        self._multiprocessing_manager.connect()

    def get_experiment_manager(self) -> ExperimentManagerProxy:
        return self._multiprocessing_manager.get_experiment_manager()  # type: ignore



from typing import Optional, Any, assert_never

from PySide6.QtCore import QObject, QModelIndex, Qt, QPersistentModelIndex
from PySide6.QtGui import QAction, QBrush, QPalette
from PySide6.QtWidgets import QMenu

from caqtus.types.expression import Expression
from caqtus.types.timelane import DigitalTimeLane
from .._colored_time_lane_model import ColoredTimeLaneModel

_DEFAULT_INDEX = QModelIndex()


class DigitalTimeLaneModel(ColoredTimeLaneModel[DigitalTimeLane]):
    # ruff: noqa: N802
    def __init__(self, name: str, parent: Optional[QObject] = None):
        lane = DigitalTimeLane([False])
        super().__init__(name, lane, parent)
        if self._brush is None:
            # If no brush is set the button will be invisible, so we pick the
            # base color from the palette as default.
            color = QPalette().text().color()
            self._brush = QBrush(color)

    def data(self, index, role=Qt.ItemDataRole.DisplayRole):
        if not index.isValid():
            return None
        value = self.lane_value(index.row())
        if role == Qt.ItemDataRole.DisplayRole:
            if isinstance(value, bool):
                return
            elif isinstance(value, Expression):
                return str(value)
            else:
                assert_never(value)
        elif role == Qt.ItemDataRole.EditRole:
            return value
        elif role == Qt.ItemDataRole.BackgroundRole:
            if isinstance(value, bool):
                if value:
                    return self._brush
        elif role == Qt.ItemDataRole.TextAlignmentRole:
            return Qt.AlignmentFlag.AlignCenter
        else:
            return super().data(index, role)

    def setData(self, index, value: Any, role: int = Qt.ItemDataRole.EditRole):
        if not index.isValid():
            return False
        if role == Qt.ItemDataRole.EditRole:
            assert isinstance(value, bool | Expression)
            return self.set_lane_value(index.row(), value)
        return False

    def insertRow(
        self, row, parent: QModelIndex | QPersistentModelIndex = _DEFAULT_INDEX
    ) -> bool:
        return self.insert_lane_value(row, False)

    def get_cell_context_actions(self, index: QModelIndex) -> list[QAction | QMenu]:
        if not index.isValid():
            return []
        cell_type_menu = QMenu("Cell type")
        value = self.lane_value(index.row())
        bool_action = cell_type_menu.addAction("on/off")
        if isinstance(value, bool):
            bool_action.setCheckable(True)
            bool_action.setChecked(True)
        else:
            bool_action.triggered.connect(
                lambda: self.setData(index, False, Qt.ItemDataRole.EditRole)
            )
        expr_action = cell_type_menu.addAction("expression")
        if isinstance(value, Expression):
            expr_action.setCheckable(True)
            expr_action.setChecked(True)
        else:
            expr_action.triggered.connect(
                lambda: self.setData(index, Expression("..."), Qt.ItemDataRole.EditRole)
            )

        return super().get_cell_context_actions(index) + [cell_type_menu]



import logging
import time
from functools import wraps
from typing import ParamSpec, TypeVar, Callable

_P = ParamSpec("_P")
_T = TypeVar("_T")


def log_duration(
    logger: logging.Logger,
) -> Callable[[Callable[_P, _T]], Callable[_P, _T]]:
    def decorator(func: Callable[_P, _T]) -> Callable[_P, _T]:
        @wraps(func)
        def wrapper(*args: _P.args, **kwargs: _P.kwargs) -> _T:
            t0 = time.time()
            result = func(*args, **kwargs)
            t1 = time.time()
            logger.debug(f"{func.__name__} took {(t1 - t0)*1e3:.2f} ms.")
            return result

        return wrapper

    return decorator



import copy
from collections.abc import Mapping, Iterable
from typing import Optional

from PySide6.QtCore import Qt, QAbstractListModel, QModelIndex
from PySide6.QtWidgets import (
    QDialog,
    QWidget,
    QVBoxLayout,
    QDialogButtonBox,
    QHBoxLayout,
    QListView,
    QStyledItemDelegate,
)

from caqtus.device import DeviceConfiguration, DeviceName
from ._add_device_dialog_ui import Ui_AddDeviceDialog
from ._device_configuration_editor import (
    DeviceConfigurationEditor,
)
from ._device_configurations_dialog_ui import Ui_DeviceConfigurationsDialog
from ._extension import CondetrolDeviceExtensionProtocol
from .._icons import get_icon

_CONFIG_ROLE = Qt.ItemDataRole.UserRole + 1
_DEFAULT_MODEL_INDEX = QModelIndex()


class DeviceConfigurationsDialog(QDialog, Ui_DeviceConfigurationsDialog):
    """A dialog for displaying and editing a collection of device configurations."""

    def __init__(
        self,
        extension: CondetrolDeviceExtensionProtocol,
        parent: Optional[QWidget] = None,
    ):
        """Initialize the dialog."""

        super().__init__(parent, Qt.WindowType.Window)
        self._configs_view = DeviceConfigurationsEditor(extension, self)
        self.add_device_dialog = AddDeviceDialog(extension, self)
        self._extension = extension
        self.setup_ui()
        self.setup_connections()

    def setup_ui(self):
        self.setupUi(self)
        layout = self.layout()
        assert isinstance(layout, QVBoxLayout)
        layout.insertWidget(0, self._configs_view, 1)
        self.add_device_button.setIcon(get_icon("plus"))
        self.remove_device_button.setIcon(get_icon("minus"))

    def setup_connections(self):
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)
        self.add_device_button.clicked.connect(self._on_add_configuration)
        self.remove_device_button.clicked.connect(
            self._configs_view.delete_selected_configuration
        )

    def _on_add_configuration(self) -> None:
        result = self.add_device_dialog.exec()
        if result == QDialog.DialogCode.Accepted:
            device_name, device_type = (
                self.add_device_dialog.device_name_line_edit.text(),
                self.add_device_dialog.device_type_combo_box.currentText(),
            )
            if not device_name:
                return
            device_configuration = self._extension.create_new_device_configuration(
                device_type
            )
            self._configs_view.add_configuration(
                DeviceName(device_name), device_configuration
            )

    def get_device_configurations(self) -> dict[DeviceName, DeviceConfiguration]:
        return self._configs_view.get_device_configurations()

    def set_device_configurations(
        self, device_configurations: Mapping[DeviceName, DeviceConfiguration]
    ) -> None:
        self._configs_view.set_device_configurations(device_configurations)


class DeviceConfigurationsEditor(QWidget):
    """A widget for displaying and editing a collection of device configurations.

    This widget displays a list of device configurations, each of which can be edited
    using a custom editor provided by the device plugin.
    """

    def __init__(
        self,
        extension: CondetrolDeviceExtensionProtocol,
        parent: Optional[QWidget] = None,
    ):
        super().__init__(parent)

        self._list_view = QListView(self)
        self._model = DeviceConfigurationModel(self)
        self._list_view.setModel(self._model)
        self._layout = QHBoxLayout(self)
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._layout.addWidget(self._list_view, 0)
        self._delegate = DeviceEditorDelegate(self._layout, extension, self)
        self._list_view.setItemDelegate(self._delegate)
        self._list_view.setEditTriggers(
            QListView.EditTrigger.DoubleClicked | QListView.EditTrigger.CurrentChanged
        )

    def set_device_configurations(
        self, device_configurations: Mapping[DeviceName, DeviceConfiguration]
    ) -> None:
        """Set the device configurations to display.

        The view will copy the configurations and store them internally, so external
        changes to the configurations will not affect the view.
        """

        self._model.set_device_configurations(device_configurations)

    def get_device_configurations(self) -> dict[DeviceName, DeviceConfiguration]:
        """Return a copy of the configurations currently displayed in the view."""

        # We need to ensure that the data in the current editor is committed before
        # we read the model data.
        w = self.editor()
        if w is not None:
            self._list_view.commitData(w)

        return self._model.get_configurations()

    def add_configuration(
        self, device_name: DeviceName, device_configuration: DeviceConfiguration
    ) -> None:
        """Add a device configuration to the view.

        The view will copy the configuration and store it internally, so external
        changes to the configuration will not affect the view.
        """

        if not isinstance(device_configuration, DeviceConfiguration):
            raise TypeError(
                f"Expected a {DeviceConfiguration}, got {type(device_configuration)}"
            )

        self._model.add_configuration(device_name, device_configuration)

    def delete_selected_configuration(self) -> None:
        """Delete the configuration that is currently selected in the view."""

        index = self._list_view.currentIndex()
        if index.isValid():
            self._model.removeRow(index.row())

    def edit(self, row: int) -> None:
        index = self._model.index(row, 0)
        self._list_view.setCurrentIndex(index)
        self._list_view.edit(index)

    def editor(self) -> Optional[DeviceConfigurationEditor]:
        w = self._list_view.indexWidget(self._list_view.currentIndex())
        assert isinstance(w, DeviceConfigurationEditor) or w is None
        return w


class DeviceEditorDelegate(QStyledItemDelegate):
    # This is a delegate that doesn't display the editor in the view cell, but instead
    # displays it in an external layout.
    def __init__(
        self,
        layout,
        extension: CondetrolDeviceExtensionProtocol,
        parent: Optional[QWidget] = None,
    ):
        super().__init__(parent)
        self._layout = layout
        self._extension = extension

    def createEditor(self, parent, option, index):
        config = index.data(_CONFIG_ROLE)
        editor = self._extension.get_device_configuration_editor(config)
        self._layout.addWidget(editor, 1)

        return editor

    def setEditorData(self, editor, index):
        # Data is already set in the editor when it is created.
        pass

    def setModelData(self, editor, model, index):
        assert isinstance(editor, DeviceConfigurationEditor)
        config = copy.deepcopy(editor.get_configuration())
        model.setData(index, config, _CONFIG_ROLE)

    def updateEditorGeometry(self, editor, option, index):
        # This is necessary to ensure that the editor is resized to fit the layout.
        editor.updateGeometry()


class AddDeviceDialog(QDialog, Ui_AddDeviceDialog):
    def __init__(
        self,
        extension: CondetrolDeviceExtensionProtocol,
        parent: Optional[QWidget] = None,
    ):
        super().__init__(parent)
        self.setup_ui(extension.available_new_configurations())

        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)

        self.device_type_combo_box.currentTextChanged.connect(
            self._on_device_type_changed
        )
        self._on_device_type_changed(self.device_type_combo_box.currentText())

    def _on_device_type_changed(self, device_type: str):
        ok_button = self.buttonBox.button(QDialogButtonBox.StandardButton.Ok)
        assert ok_button is not None
        ok_button.setEnabled(bool(device_type))

    def setup_ui(self, device_types: Iterable[str]):
        self.setupUi(self)
        for device_type in device_types:
            self.device_type_combo_box.addItem(device_type)


class DeviceConfigurationModel(QAbstractListModel):
    def __init__(self, parent: Optional[QWidget] = None):
        super().__init__(parent)

        self._device_configurations: list[tuple[DeviceName, DeviceConfiguration]] = []

    def rowCount(self, parent=QModelIndex()):
        return len(self._device_configurations)

    def data(self, index, role=Qt.ItemDataRole.DisplayRole):
        if not index.isValid():
            return None
        if role == Qt.ItemDataRole.DisplayRole:
            return self._device_configurations[index.row()][0]
        if role == _CONFIG_ROLE:
            return self._device_configurations[index.row()][1]
        return None

    def setData(self, index, value, role=Qt.ItemDataRole.EditRole) -> bool:
        if role == _CONFIG_ROLE:
            name = self._device_configurations[index.row()][0]
            self._device_configurations[index.row()] = (name, value)
            return True
        return False

    def flags(self, index):
        return (
            Qt.ItemFlag.ItemIsSelectable
            | Qt.ItemFlag.ItemIsEnabled
            | Qt.ItemFlag.ItemIsEditable
        )

    def set_device_configurations(
        self, device_configurations: Mapping[DeviceName, DeviceConfiguration]
    ) -> None:
        self.beginResetModel()
        self._device_configurations = copy.deepcopy(list(device_configurations.items()))
        self.endResetModel()

    def get_configurations(self) -> dict[DeviceName, DeviceConfiguration]:
        return dict(copy.deepcopy(self._device_configurations))

    def add_configuration(
        self, device_name: DeviceName, device_configuration: DeviceConfiguration
    ) -> None:
        self.beginInsertRows(
            QModelIndex(),
            len(self._device_configurations),
            len(self._device_configurations),
        )
        self._device_configurations.append((device_name, device_configuration))
        self.endInsertRows()

    def removeRow(self, row, parent=_DEFAULT_MODEL_INDEX):
        self.beginRemoveRows(QModelIndex(), row, row)
        del self._device_configurations[row]
        self.endRemoveRows()
        return True



import abc

import polars
from PySide6.QtWidgets import QWidget

from caqtus.gui.qtutil.qabc import QABCMeta


class DataView(QWidget, metaclass=QABCMeta):
    @abc.abstractmethod
    async def update_data(self, data: polars.DataFrame) -> None:
        raise NotImplementedError



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'exception_dialog.ui'
##
## Created by: Qt User Interface Compiler version 6.7.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import QCoreApplication, QMetaObject, Qt
from PySide6.QtWidgets import (
    QDialogButtonBox,
    QFrame,
    QLabel,
    QTreeWidget,
    QTreeWidgetItem,
    QVBoxLayout,
)


class Ui_ExceptionDialog(object):
    def setupUi(self, ExceptionDialog):
        if not ExceptionDialog.objectName():
            ExceptionDialog.setObjectName("ExceptionDialog")
        ExceptionDialog.resize(400, 300)
        self.verticalLayout = QVBoxLayout(ExceptionDialog)
        self.verticalLayout.setObjectName("verticalLayout")
        self.exception_label = QLabel(ExceptionDialog)
        self.exception_label.setObjectName("exception_label")

        self.verticalLayout.addWidget(self.exception_label)

        self.line = QFrame(ExceptionDialog)
        self.line.setObjectName("line")
        self.line.setFrameShape(QFrame.Shape.HLine)
        self.line.setFrameShadow(QFrame.Shadow.Sunken)

        self.verticalLayout.addWidget(self.line)

        self.exception_tree = QTreeWidget(ExceptionDialog)
        __qtreewidgetitem = QTreeWidgetItem()
        __qtreewidgetitem.setText(0, "1")
        self.exception_tree.setHeaderItem(__qtreewidgetitem)
        self.exception_tree.setObjectName("exception_tree")
        self.exception_tree.setHeaderHidden(True)

        self.verticalLayout.addWidget(self.exception_tree)

        self.buttonBox = QDialogButtonBox(ExceptionDialog)
        self.buttonBox.setObjectName("buttonBox")
        self.buttonBox.setOrientation(Qt.Orientation.Horizontal)
        self.buttonBox.setStandardButtons(QDialogButtonBox.StandardButton.Ok)

        self.verticalLayout.addWidget(self.buttonBox)

        self.retranslateUi(ExceptionDialog)
        self.buttonBox.accepted.connect(ExceptionDialog.accept)
        self.buttonBox.rejected.connect(ExceptionDialog.reject)

        QMetaObject.connectSlotsByName(ExceptionDialog)

    # setupUi

    def retranslateUi(self, ExceptionDialog):
        ExceptionDialog.setWindowTitle(
            QCoreApplication.translate("ExceptionDialog", "Dialog", None)
        )
        self.exception_label.setText(
            QCoreApplication.translate("ExceptionDialog", "An error occured", None)
        )

    # retranslateUi



import functools
from typing import Optional

import graphviz
import numpy as np

from ._instructions import TimedInstruction, Concatenated, Repeated, Pattern


def to_graph(instruction: TimedInstruction) -> graphviz.Digraph:
    """Convert a sequencer instruction to a graphiz graph for visualization.

    This function requires `Graphviz <https://www.graphviz.org/>`_ to be installed and
    available in the system path.

    Args:
        instruction: The instruction to convert to a graph.

    Returns:
        The graph representation of the instruction.
    """

    graph = graphviz.Digraph(graph_attr={"ordering": "in"})
    add_to_graph(instruction, graph, [0])
    return graph


def levels_to_str(levels: list[int], port: Optional[str] = None) -> str:
    value = "_".join(map(str, levels))
    if port is not None:
        value += f":{port}"
    return value


@functools.singledispatch
def add_to_graph(instr, graph, levels: list[int]) -> None:
    raise NotImplementedError(f"Cannot add {type(instr)} to graph")


@add_to_graph.register
def _(instr: Concatenated, graph, levels: list[int]):
    graph.node(levels_to_str(levels), "+", shape="circle")
    graph.attr(rank="same")
    for i, sub_instr in enumerate(instr.instructions):
        add_to_graph(sub_instr, graph, levels + [i])
        if isinstance(sub_instr, Pattern):
            port = "port"
        else:
            port = None
        graph.edge(levels_to_str(levels), levels_to_str(levels + [i], port))
    return graph


@add_to_graph.register
def _(instr: Repeated, graph, levels: list[int]):
    graph.node(levels_to_str(levels), f"*{instr.repetitions}", shape="circle")
    add_to_graph(instr.instruction, graph, levels + [0])
    if isinstance(instr.instruction, Pattern):
        port = "port"
    else:
        port = None
    graph.edge(levels_to_str(levels), levels_to_str(levels + [0], port))


@add_to_graph.register
def _(instr: Pattern, graph, levels: list[int]):
    graph.attr(shape="plain")

    rows = []

    if instr.dtype.names is None:
        cells = values_to_row(instr.array, port="port")
        rows.append(f"<TR>{'\n'.join(cells)}</TR>")
    else:
        for i, name in enumerate(instr.dtype.names):
            cells = values_to_row(instr.array[name])
            if i == 0:
                cells = [f'<TD PORT="port"><B>{name}</B></TD>'] + cells
            else:
                cells = [f"<TD><B>{name}</B></TD>"] + cells
            rows.append(f"<TR>{'\n'.join(cells)}</TR>")

    label = (
        f'<<TABLE BORDER="0" CELLBORDER="1" CELLSPACING="0">'
        f"{'\n'.join(rows)}"
        f"</TABLE>>"
    )

    graph.node(levels_to_str(levels), label, shape="plain")
    return graph


def values_to_row(values: np.ndarray, port: Optional[str] = None) -> list[str]:
    if values.dtype == np.bool_:
        values = values.astype(int)
    cells = [f"<TD>{v}</TD>" for v in values]
    if port is not None:
        if cells:
            cells[0] = f'<TD PORT="{port}">{values[0]}</TD>'
    return cells



from collections.abc import Callable

import qtawesome
from PySide6.QtGui import QFontDatabase
from PySide6.QtWidgets import QApplication

import caqtus.gui.condetrol._ressources  # noqa
from caqtus.experiment_control import ExperimentManager
from caqtus.session import ExperimentSessionMaker
from ._extension import CondetrolExtensionProtocol
from ._main_window import CondetrolMainWindow, CondetrolWindowHandler
from ..qtutil import qt_trio


# noinspection PyTypeChecker
def default_connect_to_experiment_manager() -> ExperimentManager:
    """Raise an error when trying to connect to an experiment manager."""

    error = NotImplementedError("Not implemented.")
    error.add_note(
        f"You need to provide a function to connect to the experiment "
        f"manager when initializing the main window."
    )
    error.add_note(
        "It is not possible to run sequences without connecting to an experiment "
        "manager."
    )
    raise error


class Condetrol:
    """A utility class to launch the Condetrol GUI.

    This class is a convenience wrapper around the :class:`CondetrolMainWindow` class.
    It sets up the application and launches the main window.

    See :class:`CondetrolMainWindow` for more information on the parameters.
    """

    def __init__(
        self,
        session_maker: ExperimentSessionMaker,
        extension: CondetrolExtensionProtocol,
        connect_to_experiment_manager: Callable[
            [], ExperimentManager
        ] = default_connect_to_experiment_manager,
    ):
        self.session_maker = session_maker
        app = QApplication.instance()
        if app is None:
            self.app = QApplication([])
        else:
            self.app = app
        self.app.setOrganizationName("Caqtus")
        self.app.setApplicationName("Condetrol")
        self.app.setWindowIcon(qtawesome.icon("mdi6.cactus", size=64, color="green"))  # type: ignore[reportAttributeAccessIssue]
        self.app.setStyle("Fusion")  # type: ignore[reportAttributeAccessIssue]

        QFontDatabase.addApplicationFont(":/fonts/JetBrainsMono-Regular.ttf")

        self.window = CondetrolMainWindow(
            session_maker=session_maker,
            connect_to_experiment_manager=connect_to_experiment_manager,
            extension=extension,
        )

    def run(self) -> None:
        """Launch the Condetrol GUI.

        This method will block until the GUI is closed by the user.
        """

        self.window.show()

        async def run_condetrol():
            handler = CondetrolWindowHandler(self.window, self.session_maker)
            await handler.run_async()

        qt_trio.run(run_condetrol)



import decimal
import math
from itertools import accumulate
from typing import Iterable, Sequence, NewType

Time = NewType("Time", decimal.Decimal)
"""A type for representing time in seconds.

It uses a decimal.Decimal to represent time in seconds to avoid floating point errors.
"""


ps = decimal.Decimal("1e-12")


def to_time(value: decimal.Decimal | float | str) -> Time:
    """Converts a value to a Time object.

    Args:
        value: The value to convert to a Time object, in seconds.

    Returns:
        A Time object representing the value in seconds, rounded to the picosecond.
    """

    return Time(decimal.Decimal(value).quantize(ps))


def to_time_bounds(durations: Iterable[float]) -> Sequence[Time]:
    """Converts an iterable of durations to an iterable of Time objects.

    Args:
        durations: An iterable of durations in seconds.

    Returns:
        An iterable of Time objects representing the durations in seconds.
    """

    return [to_time(duration) for duration in durations]


def get_step_bounds(step_durations: Iterable[Time]) -> Sequence[Time]:
    """Returns the time at which each step starts from their durations.

    For an iterable of step durations [d_0, d_1, ..., d_n], the step starts are
    [0, d_0, d_0 + d_1, ..., d_0 + ... + d_n]. It has one more element than the
    iterable of step durations, with the last element being the total duration.
    """

    zero = Time(decimal.Decimal(0))
    return [zero] + list((accumulate(step_durations)))


def start_tick(start_time: Time, time_step: Time) -> int:
    """Returns the included first tick index of the step starting at start_time."""

    return math.ceil(start_time / time_step)


def stop_tick(stop_time: Time, time_step: Time) -> int:
    """Returns the excluded last tick index of the step ending at stop_time."""

    return math.ceil(stop_time / time_step)


def number_ticks(start_time: Time, stop_time: Time, time_step: Time) -> int:
    """Returns the number of ticks between start_time and stop_time.

    Args:
        start_time: The start time in seconds.
        stop_time: The stop time in seconds.
        time_step: The time step in seconds.
    """

    return stop_tick(stop_time, time_step) - start_tick(start_time, time_step)



import contextlib
import datetime
import logging
from typing import Self


class DurationTimer(contextlib.AbstractContextManager):
    """A timer that measures the duration of a context."""

    def __init__(self):
        self._start_time = None
        self._end_time = None

    def __enter__(self) -> Self:
        self._start_time = datetime.datetime.now(datetime.timezone.utc)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._end_time = datetime.datetime.now(datetime.timezone.utc)

    @property
    def duration(self) -> datetime.timedelta:
        return self.end_time - self.start_time

    @property
    def duration_in_s(self) -> float:
        return self.duration.total_seconds()

    @property
    def duration_in_ms(self) -> float:
        return self.duration_in_s * 1000

    @property
    def start_time(self) -> datetime.datetime:
        if self._start_time is None:
            raise RuntimeError("Timer has not been started yet.")
        else:
            return self._start_time

    @property
    def end_time(self) -> datetime.datetime:
        if self._end_time is None:
            raise RuntimeError("Timer has not been stopped yet.")
        else:
            return self._end_time


class DurationTimerLog(DurationTimer):
    def __init__(
        self,
        logger: logging.Logger,
        message: str,
        display_start: bool = False,
        level=logging.DEBUG,
    ):
        super().__init__()
        self._logger = logger
        self._message = message
        self._display_start = display_start
        self._level = level

    def __enter__(self) -> Self:
        super().__enter__()
        if self._display_start:
            self._logger.log(self._level, f"Started: {self._message}")
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        super().__exit__(exc_type, exc_val, exc_tb)
        if exc_type is None:
            self._logger.log(
                self._level, f"Finished: {self._message} in {self.duration_in_ms} ms"
            )



from ._analog_value import (
    AnalogValue,
    NotAnalogValueError,
    is_analog_value,
    is_quantity,
    NotQuantityError,
)
from ._parameter import Parameter, is_parameter, Parameters
from ._parameter_namespace import ParameterNamespace
from ._schema import ParameterSchema

__all__ = [
    "AnalogValue",
    "NotAnalogValueError",
    "is_analog_value",
    "Parameter",
    "is_parameter",
    "Parameters",
    "ParameterNamespace",
    "is_quantity",
    "NotQuantityError",
    "ParameterSchema",
]



import anyio.lowlevel

from caqtus.utils._no_public_constructor import NoPublicConstructor


class ShotTimer(metaclass=NoPublicConstructor):
    """Gives access to pseudo-real time primitives during a shot.

    It gives the possibility to react at specific times during a shot.

    All times are relative to the start of the shot and are in seconds.
    """

    def __init__(self) -> None:
        self._start_time = anyio.current_time()

    def elapsed(self) -> float:
        """Returns the elapsed time since the start of the shot."""

        return anyio.current_time() - self._start_time

    async def wait_until(self, target_time: float) -> float:
        """Waits until a target time is reached.

        Args:
            target_time: The target time relative to the start of the shot.
                The target time can be in the past, in which case the function will
                return immediately.

        Returns:
            The duration waited for the target time to be reached.
            This duration is positive if the target time is in the future at the moment
            of the call.
            This duration is negative if the target time is in the past at the moment of
            the call.

        Raises:
            ValueError: If the target time is negative.

        Warning:
            This function is not guaranteed to be precise.
            Its accuracy depends on the underlying operating system and the event loop
            load.
        """

        if target_time < 0:
            raise ValueError("The target time must be positive.")

        absolute_target_time = self._start_time + target_time

        absolute_current_time = anyio.current_time()
        duration_to_wait = absolute_target_time - absolute_current_time

        if duration_to_wait < 0:
            # We ensure that we go through a cancellation point even if there is no
            # wait.
            await anyio.lowlevel.checkpoint()
        else:
            await anyio.sleep_until(absolute_target_time)
        return duration_to_wait



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'scatter_view.ui'
##
## Created by: Qt User Interface Compiler version 6.6.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (
    QCoreApplication,
    QDate,
    QDateTime,
    QLocale,
    QMetaObject,
    QObject,
    QPoint,
    QRect,
    QSize,
    QTime,
    QUrl,
    Qt,
)
from PySide6.QtGui import (
    QBrush,
    QColor,
    QConicalGradient,
    QCursor,
    QFont,
    QFontDatabase,
    QGradient,
    QIcon,
    QImage,
    QKeySequence,
    QLinearGradient,
    QPainter,
    QPalette,
    QPixmap,
    QRadialGradient,
    QTransform,
)
from PySide6.QtWidgets import (
    QApplication,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QSizePolicy,
    QSpacerItem,
    QToolButton,
    QVBoxLayout,
    QWidget,
)


class Ui_ScatterView(object):
    def setupUi(self, ScatterView):
        if not ScatterView.objectName():
            ScatterView.setObjectName("ScatterView")
        ScatterView.resize(400, 300)
        self.verticalLayout = QVBoxLayout(ScatterView)
        self.verticalLayout.setObjectName("verticalLayout")
        self.horizontalLayout = QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.label = QLabel(ScatterView)
        self.label.setObjectName("label")

        self.horizontalLayout.addWidget(self.label)

        self.x_line_edit = QLineEdit(ScatterView)
        self.x_line_edit.setObjectName("x_line_edit")

        self.horizontalLayout.addWidget(self.x_line_edit)

        self.horizontalSpacer = QSpacerItem(
            40, 20, QSizePolicy.Policy.Fixed, QSizePolicy.Policy.Minimum
        )

        self.horizontalLayout.addItem(self.horizontalSpacer)

        self.label_2 = QLabel(ScatterView)
        self.label_2.setObjectName("label_2")

        self.horizontalLayout.addWidget(self.label_2)

        self.y_line_edit = QLineEdit(ScatterView)
        self.y_line_edit.setObjectName("y_line_edit")

        self.horizontalLayout.addWidget(self.y_line_edit)

        self.horizontalSpacer_2 = QSpacerItem(
            40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum
        )

        self.horizontalLayout.addItem(self.horizontalSpacer_2)

        self.apply_button = QToolButton(ScatterView)
        self.apply_button.setObjectName("apply_button")

        self.horizontalLayout.addWidget(self.apply_button)

        self.settings_button = QToolButton(ScatterView)
        self.settings_button.setObjectName("settings_button")

        self.horizontalLayout.addWidget(self.settings_button)

        self.verticalLayout.addLayout(self.horizontalLayout)

        self.retranslateUi(ScatterView)

        QMetaObject.connectSlotsByName(ScatterView)

    # setupUi

    def retranslateUi(self, ScatterView):
        ScatterView.setWindowTitle(
            QCoreApplication.translate("ScatterView", "Form", None)
        )
        self.label.setText(QCoreApplication.translate("ScatterView", "x:", None))
        self.label_2.setText(QCoreApplication.translate("ScatterView", "y:", None))
        self.apply_button.setText(
            QCoreApplication.translate("ScatterView", "...", None)
        )
        self.settings_button.setText(
            QCoreApplication.translate("ScatterView", "...", None)
        )

    # retranslateUi



"""empty message

Revision ID: babf96303363
Revises: 038164d73465
Create Date: 2024-08-24 17:26:09.725307

"""

from typing import Sequence, Union

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision: str = "babf96303363"
down_revision: Union[str, None] = "038164d73465"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    op.create_table(
        "sequence.exception_tracebacks",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("sequence_id", sa.Integer(), nullable=False),
        sa.Column("content", sa.JSON(), nullable=False),
        sa.ForeignKeyConstraint(["sequence_id"], ["sequences.id"], ondelete="CASCADE"),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("sequence_id"),
    )
    op.create_index(
        op.f("ix_sequence.exception_tracebacks_sequence_id"),
        "sequence.exception_tracebacks",
        ["sequence_id"],
        unique=False,
    )


def downgrade() -> None:
    raise NotImplementedError("Downgrade is not supported")
    # ### commands auto generated by Alembic - please adjust! ###
    # op.drop_index(
    #     op.f("ix_sequence.exception_tracebacks_sequence_id"),
    #     table_name="sequence.exception_tracebacks",
    # )
    # op.drop_table("sequence.exception_tracebacks")
    # ### end Alembic commands ###



import importlib.metadata

from ._device import Device


def load_installed_devices() -> dict[str, type[Device]]:
    """Returns the device classes that are installed in the current environment.

    This function uses the entry point mechanism to find all the classes that are
    registered under the group `experiment_control.devices`.

    Returns:
        A dictionary matching the entry point name to the class indicated by the object
        reference in the entry point.
        The class referenced must be a subclass of <Device>.
        It can be an implicit subclass, as long as it has the required methods.

    Raises:
        ValueError: if two entry points are found with the same name.
        TypeError: if one of the class referenced is not a subclass of <Device>.
    """

    result: dict[str, type[Device]] = {}

    discovered_entry_points = importlib.metadata.entry_points(
        group="experiment_control.devices"
    )

    for entry_point in discovered_entry_points:
        if entry_point.name in result:
            raise ValueError(
                f"Device entry point `{entry_point.name}` cannot be used more than "
                f"once."
            )
        loaded = discovered_entry_points[entry_point.name].load()
        if not issubclass(loaded, Device):
            raise TypeError(
                f"Entry point {entry_point.name} doesn't reference a <Device> "
                f"subclass, instead got {loaded}"
            )
        result[entry_point.name] = loaded
    return result



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'dialog.ui'
##
## Created by: Qt User Interface Compiler version 6.6.1
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (
    QCoreApplication,
    QDate,
    QDateTime,
    QLocale,
    QMetaObject,
    QObject,
    QPoint,
    QRect,
    QSize,
    QTime,
    QUrl,
    Qt,
)
from PySide6.QtGui import (
    QBrush,
    QColor,
    QConicalGradient,
    QCursor,
    QFont,
    QFontDatabase,
    QGradient,
    QIcon,
    QImage,
    QKeySequence,
    QLinearGradient,
    QPainter,
    QPalette,
    QPixmap,
    QRadialGradient,
    QTransform,
)
from PySide6.QtWidgets import (
    QAbstractButton,
    QApplication,
    QDialog,
    QDialogButtonBox,
    QSizePolicy,
    QVBoxLayout,
    QWidget,
)


class Ui_ChannelOutputDialog(object):
    def setupUi(self, ChannelOutputDialog):
        if not ChannelOutputDialog.objectName():
            ChannelOutputDialog.setObjectName("ChannelOutputDialog")
        ChannelOutputDialog.resize(400, 300)
        self.verticalLayout = QVBoxLayout(ChannelOutputDialog)
        self.verticalLayout.setObjectName("verticalLayout")
        self.buttonBox = QDialogButtonBox(ChannelOutputDialog)
        self.buttonBox.setObjectName("buttonBox")
        self.buttonBox.setOrientation(Qt.Horizontal)
        self.buttonBox.setStandardButtons(QDialogButtonBox.Cancel | QDialogButtonBox.Ok)

        self.verticalLayout.addWidget(self.buttonBox)

        self.retranslateUi(ChannelOutputDialog)
        self.buttonBox.accepted.connect(ChannelOutputDialog.accept)
        self.buttonBox.rejected.connect(ChannelOutputDialog.reject)

        QMetaObject.connectSlotsByName(ChannelOutputDialog)

    # setupUi

    def retranslateUi(self, ChannelOutputDialog):
        ChannelOutputDialog.setWindowTitle(
            QCoreApplication.translate("ChannelOutputDialog", "Dialog", None)
        )

    # retranslateUi



"""Allows to pickle exceptions with their attributes and traceback.

This module is a small modification of the `tblib` package to use the `pickle.dumps`
function to pickle the exception object instead of `exception.__reduce_ex__` method.
This allows to customize the pickling behavior by using `copyreg.pickle`.
"""

import copyreg
import functools
import io
import pickle
import types
from collections.abc import Callable
from typing import Optional

import tblib
import tblib.pickling_support
import trio


def unpickle_traceback(tb_frame, tb_lineno, tb_next) -> Optional[types.TracebackType]:
    ret = object.__new__(tblib.Traceback)
    ret.tb_frame = tb_frame
    ret.tb_lineno = tb_lineno
    ret.tb_next = tb_next
    return ret.as_traceback()


def pickle_traceback(tb, *, get_locals=None):
    return unpickle_traceback, (
        tblib.Frame(tb.tb_frame, get_locals=get_locals),
        tb.tb_lineno,
        tb.tb_next and tblib.Traceback(tb.tb_next, get_locals=get_locals),
    )


def unpickle_exception(
    dump, cause, tb, context=None, suppress_context=False, notes=None
):
    inst = pickle.loads(dump)
    inst.__cause__ = cause
    inst.__traceback__ = tb
    inst.__context__ = context
    inst.__suppress_context__ = suppress_context
    if notes is not None:
        inst.__notes__ = notes
    return inst


def pickle_exception(obj):
    result = (
        unpickle_exception,
        (
            pickle.dumps(obj),
            obj.__cause__,
            obj.__traceback__,
            obj.__context__,
            obj.__suppress_context__,
            # __notes__ doesn't exist prior to Python 3.11; and even on Python 3.11 it
            # may be absent
            getattr(obj, "__notes__", None),
        ),
    )
    return result


def _get_subclasses(cls):
    # Depth-first traversal of all direct and indirect subclasses of cls
    to_visit = [cls]
    while to_visit:
        this = to_visit.pop()
        yield this
        to_visit += list(this.__subclasses__())


def ensure_exception_pickling[T, **P](func: Callable[P, T]) -> Callable[P, T]:
    """Decorator that ensures that an exception is pickled correctly."""

    @functools.wraps(func)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
        try:
            return func(*args, **kwargs)
        except BaseException as exc:
            tblib.pickling_support.install(exc)
            raise

    return wrapper


class ExceptionPickler:
    """A pickler that can pickle exceptions with their attributes and traceback.

    To pickle the exception object, the pickler uses :func:`pickle.dumps`, so it is
    possible to customize the pickling behavior by using :func:`copyreg.pickle`.
    """

    def __init__(self):
        self.dispatch_table: dict[type, Callable] = {
            types.TracebackType: functools.partial(pickle_traceback, get_locals=None)
        }
        self.install()

    def install(self) -> None:
        """Configure the pickler to be able to dump all currently defined exceptions."""

        for exception_cls in _get_subclasses(BaseException):
            self.register(exception_cls)

    def register(self, exc_type: type[BaseException]) -> None:
        self.dispatch_table.update({exc_type: pickle_exception})

    def dumps(self, obj) -> bytes:
        buffer = io.BytesIO()
        pickler = pickle.Pickler(buffer)
        pickler.dispatch_table = self.dispatch_table
        pickler.dump(obj)
        return buffer.getvalue()

    def loads(self, data: bytes):
        return pickle.loads(data)


# Trio cancelled exception cannot be pickled, so we register custom pickling functions
# for it in order to send it over the network.
def pickle_trio_cancelled_exception(exception):
    return trio.Cancelled._create, ()


copyreg.pickle(trio.Cancelled, pickle_trio_cancelled_exception)



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'sequence_widget.ui'
##
## Created by: Qt User Interface Compiler version 6.8.0
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (
    QCoreApplication,
    QDate,
    QDateTime,
    QLocale,
    QMetaObject,
    QObject,
    QPoint,
    QRect,
    QSize,
    QTime,
    QUrl,
    Qt,
)
from PySide6.QtGui import (
    QBrush,
    QColor,
    QConicalGradient,
    QCursor,
    QFont,
    QFontDatabase,
    QGradient,
    QIcon,
    QImage,
    QKeySequence,
    QLinearGradient,
    QPainter,
    QPalette,
    QPixmap,
    QRadialGradient,
    QTransform,
)
from PySide6.QtWidgets import (
    QApplication,
    QSizePolicy,
    QSplitter,
    QTabWidget,
    QUndoView,
    QVBoxLayout,
    QWidget,
)


class Ui_SequenceWidget(object):
    def setupUi(self, SequenceWidget):
        if not SequenceWidget.objectName():
            SequenceWidget.setObjectName("SequenceWidget")
        SequenceWidget.resize(742, 412)
        self.verticalLayout = QVBoxLayout(SequenceWidget)
        self.verticalLayout.setObjectName("verticalLayout")
        self.splitter = QSplitter(SequenceWidget)
        self.splitter.setObjectName("splitter")
        self.splitter.setOrientation(Qt.Orientation.Horizontal)
        self.tabWidget = QTabWidget(self.splitter)
        self.tabWidget.setObjectName("tabWidget")
        self.tabWidget.setTabPosition(QTabWidget.TabPosition.North)
        self.Constants = QWidget()
        self.Constants.setObjectName("Constants")
        self.tabWidget.addTab(self.Constants, "")
        self.iteration_tab = QWidget()
        self.iteration_tab.setObjectName("iteration_tab")
        self.tabWidget.addTab(self.iteration_tab, "")
        self.Timelanes = QWidget()
        self.Timelanes.setObjectName("Timelanes")
        self.tabWidget.addTab(self.Timelanes, "")
        self.splitter.addWidget(self.tabWidget)
        self.undoView = QUndoView(self.splitter)
        self.undoView.setObjectName("undoView")
        self.splitter.addWidget(self.undoView)

        self.verticalLayout.addWidget(self.splitter)

        self.retranslateUi(SequenceWidget)

        self.tabWidget.setCurrentIndex(2)

        QMetaObject.connectSlotsByName(SequenceWidget)

    # setupUi

    def retranslateUi(self, SequenceWidget):
        SequenceWidget.setWindowTitle(
            QCoreApplication.translate("SequenceWidget", "Form", None)
        )
        self.tabWidget.setTabText(
            self.tabWidget.indexOf(self.Constants),
            QCoreApplication.translate("SequenceWidget", "Constants", None),
        )
        self.tabWidget.setTabText(
            self.tabWidget.indexOf(self.iteration_tab),
            QCoreApplication.translate("SequenceWidget", "Iteration", None),
        )
        self.tabWidget.setTabText(
            self.tabWidget.indexOf(self.Timelanes),
            QCoreApplication.translate("SequenceWidget", "Shot", None),
        )

    # retranslateUi



from __future__ import annotations

from typing import Optional, assert_never

import attrs
import numpy as np
import pyqtgraph
from PySide6.QtWidgets import QDialog
from PySide6.QtWidgets import QWidget

from caqtus.device import DeviceName
from caqtus.session import ExperimentSessionMaker
from caqtus.session._data_id import DataId
from caqtus.session._shot_id import ShotId
from caqtus.types.data import DataLabel
from caqtus.types.image import ImageLabel, Image, is_image
from caqtus.types.recoverable_exceptions import InvalidTypeError
from caqtus.utils import serialization
from .image_view_dialog_ui import Ui_ImageViewDialog
from ..single_shot_view import ShotView


@attrs.define
class ImageViewState:
    """Contains the state of a view for an image.

    Attributes:
        camera_name: The name of the camera device from which to fetch the image.
        image: The name of the image to fetch.
        background: The name of the background image to fetch.
        colormap: The name of the colormap to use.
        levels: The minimum and maximum values to display.
    """

    camera_name: DeviceName
    image: ImageLabel
    background: Optional[ImageLabel] = None
    colormap: Optional[str] = None
    levels: Optional[tuple[float, float]] = None


class ImageView(ShotView, pyqtgraph.ImageView):
    """A view to display an image for a shot."""

    def __init__(
        self,
        state: ImageViewState,
        session_maker: ExperimentSessionMaker,
        parent: Optional[QWidget] = None,
    ):
        super().__init__(parent)
        self._state = state
        self.set_state(state)
        self._session_maker = session_maker
        self.getHistogramWidget().item.sigLevelChangeFinished.connect(
            self._on_levels_changed
        )

    def get_state(self) -> ImageViewState:
        return self._state

    def set_state(self, state: ImageViewState) -> None:
        if state.colormap is not None:
            colormap = pyqtgraph.colormap.get(state.colormap, source="matplotlib")
            self.setColorMap(colormap)
        if state.levels is not None:
            self.setLevels(*state.levels)

    async def display_shot(self, shot: ShotId) -> None:
        image_label = ImageLabel(
            DataLabel(f"{self._state.camera_name}\\{self._state.image}")
        )
        background_label = (
            ImageLabel(
                DataLabel(f"{self._state.camera_name}\\{self._state.background}")
            )
            if self._state.background is not None
            else None
        )

        image = await load_image(
            shot, image_label, background_label, self._session_maker
        )
        self.set_image(image)

    def set_image(self, image: Image) -> None:
        match self._state.levels:
            case None:
                auto_range = True
                levels = None
                auto_histogram_range = True
            case (min_, max_):
                auto_range = False
                auto_histogram_range = False
                levels = (min_, max_)
            case _:
                assert_never(self._state.levels)
        self.setImage(
            image[::, ::-1],
            autoRange=auto_range,
            levels=levels,
            autoHistogramRange=auto_histogram_range,
        )

    def _on_levels_changed(self) -> None:
        if self._state.levels is not None:
            self._state.levels = (
                self.getLevels()  # pyright: ignore[reportAttributeAccessIssue]
            )


async def load_image(
    shot: ShotId,
    image_label: ImageLabel,
    background_label: Optional[ImageLabel],
    session_make: ExperimentSessionMaker,
) -> Image[np.floating]:
    async with session_make.async_session() as session:
        image = await session.sequences.get_shot_data_by_label(
            DataId(shot, image_label)
        )
        if not is_image(image):
            raise InvalidTypeError(f"Data {image_label} is not an image.")
        image = image.astype(float)
        if background_label is not None:
            background = await session.sequences.get_shot_data_by_label(
                DataId(shot, background_label)
            )
            if not is_image(background):
                raise InvalidTypeError(f"Data {background_label} is not an image.")
            background = background.astype(float)
            image = image - background
    return image


class ImageViewDialog(QDialog, Ui_ImageViewDialog):
    def __init__(self):
        super().__init__()
        self.setupUi(self)


def create_image_view() -> tuple[str, ImageViewState]:
    dialog = ImageViewDialog()
    result = dialog.exec()
    if result == QDialog.DialogCode.Accepted:
        name = dialog.view_name_line_edit.text()
        state = ImageViewState(
            camera_name=DeviceName(dialog.camera_name_line_edit.text()),
            image=ImageLabel(DataLabel(dialog.image_line_edit.text())),
            background=(
                ImageLabel(DataLabel(text))
                if (text := dialog.background_line_edit.text())
                else None
            ),
        )
        return name, serialization.unstructure(state)

    else:
        raise ValueError("Dialog was cancelled.")



from typing import Protocol

from caqtus.device.configuration.serializer import DeviceConfigJSONSerializerProtocol
from caqtus.experiment_control.device_manager_extension import (
    DeviceManagerExtensionProtocol,
)
from caqtus.gui.condetrol._extension import CondetrolExtensionProtocol
from caqtus.types.timelane._serializer import TimeLaneSerializerProtocol


class CaqtusExtensionProtocol(Protocol):
    @property
    def condetrol_extension(self) -> CondetrolExtensionProtocol: ...

    @property
    def device_configurations_serializer(
        self,
    ) -> DeviceConfigJSONSerializerProtocol: ...

    @property
    def time_lane_serializer(self) -> TimeLaneSerializerProtocol: ...

    @property
    def device_manager_extension(self) -> DeviceManagerExtensionProtocol: ...



from __future__ import annotations

import functools
from collections.abc import Iterable

from PySide6 import QtWidgets

from ._editor_builder import EditorFactory, ValueEditor


def build_literal_editor(*values) -> EditorFactory[LiteralEditor]:
    """Build an editor for a predefined set of values.

    The editor will be a combo box proposing the given values.

    Args:
        values: The values to propose in the combo box.
            They must be hashable and convertible to string.
    """

    if not values:
        raise ValueError("At least one value must be provided")
    return functools.partial(LiteralEditor, values)


class LiteralEditor(ValueEditor):
    def __init__(self, values: Iterable) -> None:
        if not values:
            raise ValueError("At least one value must be provided")
        self.combo_box = QtWidgets.QComboBox()
        self.value_to_index = {}
        self.index_to_value = {}
        for index, value in enumerate(values):
            self.combo_box.addItem(str(value))
            self.value_to_index[value] = index
            self.index_to_value[index] = value
        self.combo_box.setCurrentIndex(0)

    def set_value(self, value) -> None:
        index = self.value_to_index[value]
        self.combo_box.setCurrentIndex(index)

    def read_value(self):
        return self.index_to_value[self.combo_box.currentIndex()]

    def set_editable(self, editable: bool) -> None:
        self.combo_box.setEnabled(editable)

    @property
    def widget(self) -> QtWidgets.QComboBox:
        return self.combo_box



import abc

from PySide6.QtWidgets import QWidget


class ValueEditor[T](abc.ABC):
    """Allows to edit a value of type T."""

    @abc.abstractmethod
    def set_value(self, value: T) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    def read_value(self) -> T:
        """Return the current value displayed in the editor."""

        raise NotImplementedError

    @abc.abstractmethod
    def set_editable(self, editable: bool) -> None:
        """Set whether the editor is editable or not.

        When initialized, the editor is editable.
        """

        raise NotImplementedError

    @property
    @abc.abstractmethod
    def widget(self) -> QWidget:
        """Return the widget that allows to edit the value."""

        raise NotImplementedError



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'add_lane_dialog.ui'
##
## Created by: Qt User Interface Compiler version 6.6.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (
    QCoreApplication,
    QDate,
    QDateTime,
    QLocale,
    QMetaObject,
    QObject,
    QPoint,
    QRect,
    QSize,
    QTime,
    QUrl,
    Qt,
)
from PySide6.QtGui import (
    QBrush,
    QColor,
    QConicalGradient,
    QCursor,
    QFont,
    QFontDatabase,
    QGradient,
    QIcon,
    QImage,
    QKeySequence,
    QLinearGradient,
    QPainter,
    QPalette,
    QPixmap,
    QRadialGradient,
    QTransform,
)
from PySide6.QtWidgets import (
    QAbstractButton,
    QApplication,
    QComboBox,
    QDialog,
    QDialogButtonBox,
    QFormLayout,
    QLabel,
    QLineEdit,
    QSizePolicy,
    QVBoxLayout,
    QWidget,
)


class Ui_AddLaneDialog(object):
    def setupUi(self, AddLaneDialog):
        if not AddLaneDialog.objectName():
            AddLaneDialog.setObjectName("AddLaneDialog")
        AddLaneDialog.resize(400, 300)
        self.verticalLayout = QVBoxLayout(AddLaneDialog)
        self.verticalLayout.setObjectName("verticalLayout")
        self.formLayout = QFormLayout()
        self.formLayout.setObjectName("formLayout")
        self.lane_name_label = QLabel(AddLaneDialog)
        self.lane_name_label.setObjectName("lane_name_label")

        self.formLayout.setWidget(0, QFormLayout.LabelRole, self.lane_name_label)

        self.lane_name_line_edit = QLineEdit(AddLaneDialog)
        self.lane_name_line_edit.setObjectName("lane_name_line_edit")

        self.formLayout.setWidget(0, QFormLayout.FieldRole, self.lane_name_line_edit)

        self.laneTypeLabel = QLabel(AddLaneDialog)
        self.laneTypeLabel.setObjectName("laneTypeLabel")

        self.formLayout.setWidget(1, QFormLayout.LabelRole, self.laneTypeLabel)

        self.lane_type_combobox = QComboBox(AddLaneDialog)
        self.lane_type_combobox.setObjectName("lane_type_combobox")

        self.formLayout.setWidget(1, QFormLayout.FieldRole, self.lane_type_combobox)

        self.verticalLayout.addLayout(self.formLayout)

        self.buttonBox = QDialogButtonBox(AddLaneDialog)
        self.buttonBox.setObjectName("buttonBox")
        self.buttonBox.setOrientation(Qt.Horizontal)
        self.buttonBox.setStandardButtons(QDialogButtonBox.Cancel | QDialogButtonBox.Ok)

        self.verticalLayout.addWidget(self.buttonBox)

        self.retranslateUi(AddLaneDialog)
        self.buttonBox.accepted.connect(AddLaneDialog.accept)
        self.buttonBox.rejected.connect(AddLaneDialog.reject)

        QMetaObject.connectSlotsByName(AddLaneDialog)

    # setupUi

    def retranslateUi(self, AddLaneDialog):
        AddLaneDialog.setWindowTitle(
            QCoreApplication.translate("AddLaneDialog", "Dialog", None)
        )
        self.lane_name_label.setText(
            QCoreApplication.translate("AddLaneDialog", "Lane name", None)
        )
        self.lane_name_line_edit.setPlaceholderText(
            QCoreApplication.translate("AddLaneDialog", "Name", None)
        )
        self.laneTypeLabel.setText(
            QCoreApplication.translate("AddLaneDialog", "Lane type", None)
        )

    # retranslateUi



"""Contains code to run a sequence."""

from ._sequence_manager import (
    SequenceManager,
    ShotRetryConfig,
)
from ._sequence_manager import run_sequence
from .shot_timing import ShotTimer

__all__ = [
    "SequenceManager",
    "ShotRetryConfig",
    "ShotTimer",
    "run_sequence",
]



from __future__ import annotations

from typing import Optional

import anyio.to_thread
import attrs
import numpy as np
import polars
import pyqtgraph
import qtawesome
from PySide6.QtCore import QStringListModel, QTimer, Qt
from PySide6.QtGui import QPen, QFont
from PySide6.QtWidgets import QWidget, QDialog, QCompleter, QBoxLayout
from caqtus.analysis.stats import compute_stats_average, get_nominal_value, get_error
from caqtus.analysis.units import extract_unit
from caqtus.gui.graphplot.views.view import DataView
from caqtus.gui.qtutil import temporary_widget

from .error_bar_view_ui import Ui_ErrorBarView
from .settings_dialog_ui import Ui_SettingsDialog


class ErrorBarView(DataView, Ui_ErrorBarView):
    def __init__(self, parent: Optional[QWidget] = None):
        super().__init__(parent)

        self.setupUi(self)

        self.settings_button.setIcon(qtawesome.icon("mdi6.cog"))
        self.settings_button.clicked.connect(self.on_settings_button_clicked)

        self.columns_model = QStringListModel(self)
        layout = self.layout()
        assert isinstance(layout, QBoxLayout)
        self._layout = layout

        self.plot: Optional[ErrorBarPlot] = None

    def on_settings_button_clicked(self) -> None:
        with temporary_widget(
            SettingsDialog("View settings...", self.columns_model, self)
        ) as dialog:
            ok = dialog.exec_()
            settings = dialog.get_settings()
        if ok:
            if self.plot is not None:
                self.plot.deleteLater()
            if settings.hue_column is None:
                self.plot = ErrorBarPlot(settings.x_column, settings.y_column, self)
                self._layout.insertWidget(1, self.plot)
            else:
                raise NotImplementedError("Hue column not supported yet.")

    async def update_data(self, data: polars.DataFrame) -> None:
        column_names = data.columns
        if column_names != self.columns_model.stringList():
            # We only reset the completer if the columns have actually changed,
            # otherwise we would reset the user input every time the data is updated.
            self.columns_model.setStringList(data.columns)
        if self.plot is not None:
            await self.plot.update_data(data)


class SettingsDialog(QDialog, Ui_SettingsDialog):
    def __init__(
        self,
        title: str,
        columns_model: QStringListModel,
        parent: Optional[QWidget] = None,
    ):
        super().__init__(parent)

        self.setupUi(self)
        self.setWindowTitle(title)
        self.ok_button.clicked.connect(self.accept)
        self.columns_completer = QCompleter(columns_model, self)
        self.x_line_edit.setCompleter(self.columns_completer)
        self.y_line_edit.setCompleter(self.columns_completer)
        self.hue_line_edit.setCompleter(self.columns_completer)

    def get_settings(self) -> ErrorBarSettings:
        return ErrorBarSettings(
            x_column=self.x_line_edit.text(),
            y_column=self.y_line_edit.text(),
            hue_column=self.hue_line_edit.text() or None,
        )


@attrs.define
class ErrorBarSettings:
    x_column: str
    y_column: str
    hue_column: Optional[str] = None


class ErrorBarPlot(pyqtgraph.PlotWidget):
    def __init__(self, x_column: str, y_column: str, parent: Optional[QWidget] = None):
        super().__init__(parent, background="white")
        black_pen = QPen(Qt.GlobalColor.black)
        self.getAxis("bottom").setTextPen(black_pen)
        self.getAxis("left").setTextPen(black_pen)
        font = QFont()
        font.setPointSize(15)
        self.getAxis("bottom").setTickFont(font)
        self.getAxis("left").setTickFont(font)
        self.getAxis("bottom").label.setFont(font)
        self.getAxis("left").label.setFont(font)
        self.enableAutoRange()

        self.error_bar_item = pyqtgraph.ErrorBarItem()
        self.scatter_plot = pyqtgraph.ScatterPlotItem()
        plot_item = self.getPlotItem()
        assert plot_item is not None
        self.plot_item = plot_item
        self.plot_item.addItem(self.error_bar_item)
        self.plot_item.addItem(self.scatter_plot)
        self.x_column = x_column
        self.y_column = y_column

    def clear(self) -> None:
        self.error_bar_item.setData(x=[], y=[], height=[])
        self.plot_item.setLabel("bottom", self.x_column)
        self.plot_item.setLabel("left", self.y_column)

    async def update_data(self, data: polars.DataFrame) -> None:
        if data.is_empty():
            self.clear()
            return
        average = await anyio.to_thread.run_sync(
            compute_stats_average, data, [self.y_column], [self.x_column]
        )
        x_magnitudes, x_unit = extract_unit(average[self.x_column])
        y_magnitudes, y_unit = extract_unit(average[self.y_column])
        x = np.array(x_magnitudes)
        y = np.array(get_nominal_value(y_magnitudes))

        self.error_bar_item.setData(
            x=x,
            y=y,
            height=np.array(get_error(y_magnitudes) * 2),
        )
        self.scatter_plot.setData(x=x, y=y)

        if x_unit:
            x_label = f"{self.x_column} [{x_unit}]"
        else:
            x_label = self.x_column
        self.plot_item.setLabel("bottom", x_label)

        if y_unit:
            y_label = f"{self.y_column} [{y_unit}]"
        else:
            y_label = self.y_column
        self.plot_item.setLabel("left", y_label)



"""Contains classes for loading data for a shot."""

from ._combinable_importers import CombinableLoader, join
from ._load_image_count import LoadImageCount
from ._load_parameters import LoadShotParameters
from ._load_shot_id import LoadShotId
from ._load_shot_info import LoadShotTime
from ._shot_data import ShotImporter, DataImporter

__all__ = [
    "CombinableLoader",
    "LoadShotParameters",
    "LoadShotId",
    "LoadShotTime",
    "ShotImporter",
    "join",
    "DataImporter",
    "LoadImageCount",
]



"""Used to compile user-friendly parameters into low-level device parameters."""

from caqtus.types.units.unit_namespace import units
from . import lane_compilation, timed_instructions, timing
from ._device_compiler import DeviceCompiler, DeviceNotUsedException
from .compilation_contexts import ShotContext, SequenceContext

__all__ = [
    "units",
    "ShotContext",
    "SequenceContext",
    "DeviceCompiler",
    "DeviceNotUsedException",
    "lane_compilation",
    "timed_instructions",
    "timing",
]



"""Contains functions to compile lanes into raw values."""

from ._compile_analog_lane import compile_analog_lane, DimensionedSeries
from ._compile_digital_lane import compile_digital_lane

__all__ = ["compile_digital_lane", "compile_analog_lane", "DimensionedSeries"]



from typing import Protocol, runtime_checkable

from ..device_configuration_editors._extension import CondetrolDeviceExtensionProtocol
from ..timelanes_editor.extension import CondetrolLaneExtensionProtocol


@runtime_checkable
class CondetrolExtensionProtocol(Protocol):
    """Defines the operations an extension must implement to be used by Condetrol."""

    @property
    def device_extension(self) -> CondetrolDeviceExtensionProtocol: ...

    @property
    def lane_extension(self) -> CondetrolLaneExtensionProtocol: ...



from __future__ import annotations

import datetime
from typing import Mapping, Any

import attrs

from caqtus.device import DeviceName
from caqtus.types._parameter_namespace import VariableNamespace
from caqtus.types.data import DataLabel, Data


@attrs.frozen(order=True)
class ShotParameters:
    """Holds information necessary to compile a shot."""

    index: int
    parameters: VariableNamespace = attrs.field(eq=False)


@attrs.frozen(order=True)
class DeviceParameters:
    """Holds information necessary to run a shot."""

    index: int
    shot_parameters: VariableNamespace = attrs.field(eq=False)
    device_parameters: Mapping[DeviceName, Mapping[str, Any]] = attrs.field(eq=False)
    timeout: float = attrs.field()


@attrs.frozen(order=True)
class ShotData:
    """Holds information necessary to store a shot."""

    index: int
    start_time: datetime.datetime = attrs.field(eq=False)
    end_time: datetime.datetime = attrs.field(eq=False)
    variables: VariableNamespace = attrs.field(eq=False)
    data: Mapping[DataLabel, Data] = attrs.field(eq=False)



from caqtus.types.iteration import StepsConfiguration
from caqtus.types.iteration._step_context import StepContext
from caqtus.types.parameter import Parameter
from .shots_manager import ShotScheduler


async def execute_steps(
    steps: StepsConfiguration,
    initial_context: StepContext[Parameter],
    shot_scheduler: ShotScheduler,
):
    """Execute a sequence of steps on the experiment.

    This method will recursively execute each step in the sequence passed as
    argument and scheduling the shots when encountering a shot step.
    """

    for context in steps.walk(initial_context):
        await shot_scheduler.schedule_shot(context.variables)



"""Allows to create user-defined components to use on an experiment."""

from ._experiment import Experiment, upgrade_database, stamp_database
from .device_extension import DeviceExtension
from .time_lane_extension import TimeLaneExtension

__all__ = [
    "Experiment",
    "DeviceExtension",
    "TimeLaneExtension",
    "upgrade_database",
    "stamp_database",
]



"""This package provides a widget to edit the configuration of a camera."""

from ._editor import CameraConfigurationEditor, RectangularROIEditor

__all__ = ["CameraConfigurationEditor", "RectangularROIEditor"]



from __future__ import annotations

import functools
from typing import Optional

import attrs
from PySide6.QtCore import (
    QAbstractTableModel,
    QObject,
    QModelIndex,
    Qt,
    QSize,
    QPersistentModelIndex,
)
from PySide6.QtGui import QAction, QUndoStack, QUndoCommand
from PySide6.QtWidgets import QMenu

from caqtus.types.expression import Expression
from caqtus.types.timelane import TimeLanes, TimeLane
from ._time_lane_model import TimeLaneModel
from ._time_step_model import (
    TimeStepNameModel,
    TimeStepDurationModel,
)
from .extension import CondetrolLaneExtensionProtocol

_DEFAULT_MODEL_INDEX = QModelIndex()


class TimeLanesModel(QAbstractTableModel):
    """Table model for time lanes of a sequence.

    Each column represents a step in the shot.
    The first row contains the names of the steps.
    The second row contains the durations of the steps.
    The following rows contain the lanes of the sequence.

    This model has an undo stack when actions are performed.
    """

    # Ignore some lint rules for this class as PySide6 models have a lot of camelCase
    # methods.
    # ruff: noqa: N802
    def __init__(
        self,
        extension: CondetrolLaneExtensionProtocol,
        parent: Optional[QObject] = None,
    ):
        super().__init__(parent)
        self.undo_stack = QUndoStack()
        self.undo_stack.setClean()
        self._step_names_model = TimeStepNameModel(self.undo_stack, self)
        self._step_durations_model = TimeStepDurationModel(self.undo_stack, self)
        self._lane_models: list[TimeLaneModel] = []
        self._extension = extension

        self._step_names_model.dataChanged.connect(self._on_step_names_data_changed)
        self._step_durations_model.dataChanged.connect(
            self._on_step_durations_data_changed
        )
        self._read_only = False

    def set_read_only(self, read_only: bool) -> None:
        self._read_only = read_only

    def is_read_only(self) -> bool:
        return self._read_only

    def has_uncommitted_edits(self) -> bool:
        """Indicate if the user has edited data since calling :meth:`commit_edits`.

        This value is orthogonal to :meth:`is_read_only` as the model can be read only
        and still have uncommitted edits.
        """

        return not self.undo_stack.isClean()

    def commit_edits(self) -> None:
        """Marks the model has having no uncommitted edits.

        Typically, one would call this method after saving the result of
        :meth:`get_timelanes`.
        """

        self.undo_stack.setClean()

    def get_timelanes(self) -> TimeLanes:
        """Return a copy of the lanes currently in the model."""

        return TimeLanes(
            step_names=self._step_names_model.get_names(),
            step_durations=self._step_durations_model.get_duration(),
            lanes={
                get_lane_model_name(model): model.get_lane()
                for model in self._lane_models
            },
        )

    def cell_index(self, lane_index: int, step: int) -> QModelIndex:

        return self.index(lane_index + 2, step)

    def _on_step_names_data_changed(
        self,
        top_left: QModelIndex,
        bottom_right: QModelIndex,
        roles: list[Qt.ItemDataRole],
    ):
        self.dataChanged.emit(
            self.index(0, top_left.row()), self.index(0, bottom_right.row())
        )

    def _on_step_durations_data_changed(
        self,
        top_left: QModelIndex,
        bottom_right: QModelIndex,
        roles: list[Qt.ItemDataRole],
    ):
        self.dataChanged.emit(
            self.index(1, top_left.row()), self.index(1, bottom_right.row())
        )

    def set_timelanes(self, timelanes: TimeLanes):
        """Set the lanes of the model.

        Calling this method clears the undo stack of the model.
        """

        # Don't check if read only, because we need to update the content of the editor
        # even if it is readonly when swapping sequences.
        new_models = []
        for name, lane in timelanes.lanes.items():
            lane_model = self._create_lane_model(name, lane)
            new_models.append(lane_model)

        self.beginResetModel()
        self.undo_stack.clear()
        self._step_names_model.set_names(timelanes.step_names)
        self._step_durations_model.set_durations(timelanes.step_durations)
        self._lane_models.clear()
        self._lane_models.extend(new_models)
        self.endResetModel()

    def set_timelanes_with_undo(self, timelanes: TimeLanes, message: str) -> bool:
        if self._read_only:
            return False
        new_models = []
        for name, lane in timelanes.lanes.items():
            lane_model = self._create_lane_model(name, lane)
            new_models.append(lane_model)
        self.undo_stack.push(
            self._SetTimeLanesCommand(
                self,
                list(self._lane_models),
                list(new_models),
                self._step_names_model.get_names(),
                list(timelanes.step_names),
                list(self._step_durations_model.get_duration()),
                list(timelanes.step_durations),
                message,
            )
        )
        return True

    @attrs.define(slots=False)
    class _SetTimeLanesCommand(QUndoCommand):
        model: TimeLanesModel
        old_models: list[TimeLaneModel]
        new_models: list[TimeLaneModel]
        old_names: list[str]
        new_names: list[str]
        old_durations: list[Expression]
        new_durations: list[Expression]
        message: str

        def __attrs_post_init__(self):
            super().__init__(self.message)

        def redo(self):
            self.model.beginResetModel()
            self.model._step_names_model.set_names(self.new_names)
            self.model._step_durations_model.set_durations(self.new_durations)
            self.model._lane_models.clear()
            self.model._lane_models.extend(self.new_models)
            self.model.endResetModel()

        def undo(self):
            self.model.beginResetModel()
            self.model._step_names_model.set_names(self.old_names)
            self.model._step_durations_model.set_durations(self.old_durations)
            self.model._lane_models.clear()
            self.model._lane_models.extend(self.old_models)
            self.model.endResetModel()

    def _create_lane_model(self, name: str, lane: TimeLane) -> TimeLaneModel:
        lane_model = self._extension.get_lane_model(lane, name)
        lane_model.setParent(self)
        lane_model.set_lane(lane)
        lane_model.set_undo_stack(self.undo_stack)
        lane_model.dataChanged.connect(
            # For some reason, functools.partial does not work here, but lambda does.
            # functools.partial(
            #     self.on_lane_model_data_changed, lane_model=lane_model
            # )
            lambda top_left, bottom_right: self._on_lane_model_data_changed(
                top_left, bottom_right, lane_model
            )
        )
        lane_model.modelReset.connect(
            functools.partial(self._on_lane_model_reset, lane_model=lane_model)
        )
        lane_model.headerDataChanged.connect(
            functools.partial(self._on_lane_header_data_changed, lane_model=lane_model)
        )
        return lane_model

    def _on_lane_model_reset(self, lane_model: TimeLaneModel):
        lane_index = self._lane_models.index(lane_model)
        self.dataChanged.emit(
            self.index(lane_index + 2, 0),
            self.index(lane_index + 2, self.columnCount() - 1),
        )

    def _on_lane_model_data_changed(
        self,
        top_left: QModelIndex,
        bottom_right: QModelIndex,
        lane_model: TimeLaneModel,
    ):
        lane_index = self._lane_models.index(lane_model)
        self.dataChanged.emit(
            self.index(lane_index + 2, top_left.row()),
            self.index(lane_index + 2, bottom_right.row()),
        )

    def _on_lane_header_data_changed(
        self,
        orientation: Qt.Orientation,
        first: int,
        last: int,
        lane_model: TimeLaneModel,
    ):
        lane_index = self._lane_models.index(lane_model)
        if orientation == Qt.Orientation.Horizontal:
            self.headerDataChanged.emit(
                Qt.Orientation.Vertical,
                lane_index + 2,
                lane_index + 2,
            )

    def get_lane(self, index: int) -> TimeLane:
        return self._lane_models[index].get_lane()

    def get_lane_name(self, index: int) -> str:
        return get_lane_model_name(self._lane_models[index])

    def number_steps(self) -> int:
        count = self._step_names_model.rowCount()
        assert count == self._step_durations_model.rowCount()
        return count

    def get_step_name(self, step: int) -> str:
        name = self._step_names_model.data(
            self._step_names_model.index(step), Qt.ItemDataRole.DisplayRole
        )
        assert isinstance(name, str)
        return name

    def columnCount(
        self, parent: QModelIndex | QPersistentModelIndex = _DEFAULT_MODEL_INDEX
    ) -> int:
        return self.number_steps()

    def rowCount(
        self, parent: QModelIndex | QPersistentModelIndex = _DEFAULT_MODEL_INDEX
    ) -> int:
        return len(self._lane_models) + 2

    def data(self, index, role: int = Qt.ItemDataRole.DisplayRole):
        if not index.isValid():
            return None
        return self._map_to_source(index).data(role)

    def setData(self, index, value, role: int = Qt.ItemDataRole.EditRole) -> bool:
        if self._read_only:
            return False
        if not index.isValid():
            return False
        if role != Qt.ItemDataRole.EditRole:
            return False
        if index.row() == 0:
            return self._step_names_model.setData(
                self._step_names_model.index(index.column()), value
            )
        elif index.row() == 1:
            return self._step_durations_model.setData(
                self._step_durations_model.index(index.column()), value
            )
        else:
            return self._lane_models[index.row() - 2].setData(
                self._map_to_source(index), value
            )

    def _set_data(self, value, row: int, column: int) -> None:
        index = self.index(row, column)
        mapped_index = self._map_to_source(index)
        result = mapped_index.model().setData(
            mapped_index, value, Qt.ItemDataRole.EditRole
        )
        assert result

    class _SetValueCommand(QUndoCommand):
        def __init__(self, model: TimeLanesModel, new_value, row: int, column: int):
            self.previous_value = model.data(
                model.index(row, column), Qt.ItemDataRole.EditRole
            )
            if row == 0:
                msg = (
                    f"change step {column} name from <{self.previous_value}> to "
                    f"<{new_value}>"
                )
            elif row == 1:
                msg = (
                    f"change step {column} duration from <{self.previous_value}> to "
                    f"<{new_value}>"
                )
            else:
                msg = (
                    f"change value of lane {model.get_lane_name(row-2)} from "
                    f"<{self.previous_value}> to <{new_value}> for step {column}"
                )

            super().__init__(msg)
            self.model = model
            self.new_value = new_value
            self.row = row
            self.column = column

        def redo(self) -> None:
            self.model._set_data(self.new_value, self.row, self.column)

        def undo(self) -> None:
            self.model._set_data(self.previous_value, self.row, self.column)

    def flags(self, index) -> Qt.ItemFlag:
        if not index.isValid():
            return Qt.ItemFlag.NoItemFlags
        mapped_index = self._map_to_source(index)
        flags = mapped_index.model().flags(mapped_index)
        if self._read_only:
            flags &= ~Qt.ItemFlag.ItemIsEditable
            flags &= ~Qt.ItemFlag.ItemIsDropEnabled
        return flags

    def headerData(
        self,
        section: int,
        orientation: Qt.Orientation,
        role: int = Qt.ItemDataRole.DisplayRole,
    ):
        if orientation == Qt.Orientation.Horizontal:
            if role == Qt.ItemDataRole.DisplayRole:
                return f"Step {section}"
        elif orientation == Qt.Orientation.Vertical:
            if section == 0:
                return self._step_names_model.headerData(
                    0, Qt.Orientation.Horizontal, role
                )
            elif section == 1:
                return self._step_durations_model.headerData(
                    0, Qt.Orientation.Horizontal, role
                )
            else:
                return self._lane_models[section - 2].headerData(
                    0, Qt.Orientation.Horizontal, role
                )

    def insertColumn(
        self,
        column: int,
        parent: QModelIndex | QPersistentModelIndex = _DEFAULT_MODEL_INDEX,
    ) -> bool:
        """Add a new time step at the requested column.

        If performed, pushes the action on the model undo stack.

        Returns:
            False if no action was performed if the model is read only or if the column
            index is not valid.
            True otherwise if the action was performed.
        """

        if self._read_only:
            return False
        if not (0 <= column <= self.columnCount()):
            return False
        self.undo_stack.beginMacro(f"insert step {column}")
        self.undo_stack.push(self._BeginInsertColumnCommand(self, column))
        self._step_names_model.insertRow(column)
        self._step_durations_model.insertRow(column)
        for model in self._lane_models:
            model.insertRow(column)
        self.undo_stack.push(self._EndInsertColumnCommand(self, column))
        self.undo_stack.endMacro()
        return True

    @attrs.frozen(slots=False)
    class _BeginInsertColumnCommand(QUndoCommand):
        model: TimeLanesModel
        column: int

        def __attrs_post_init__(self):
            super().__init__(f"begin insert column {self.column}")

        def redo(self) -> None:
            assert not self.model._read_only

            self.model.beginInsertColumns(QModelIndex(), self.column, self.column)

        def undo(self) -> None:
            assert not self.model._read_only
            self.model.endRemoveColumns()

    @attrs.frozen(slots=False)
    class _EndInsertColumnCommand(QUndoCommand):
        model: TimeLanesModel
        column: int

        def __attrs_post_init__(self):
            super().__init__("end insert column {self.column}")

        def redo(self) -> None:
            assert not self.model._read_only
            self.model.endInsertColumns()

        def undo(self) -> None:
            assert not self.model._read_only
            self.model.beginRemoveColumns(QModelIndex(), self.column, self.column)

    def removeColumn(
        self, column, parent: QModelIndex | QPersistentModelIndex = _DEFAULT_MODEL_INDEX
    ) -> bool:
        """Remove a step from the model.

        If an action was performed, pushes the action on the model undo stack.

        Returns:
            False if no action was performed if the model is read only or if the column
            index is not valid.
            True otherwise.
        """

        if self._read_only:
            return False
        if not (0 <= column < self.columnCount()):
            return False
        self.undo_stack.beginMacro(f"remove step {column}")
        self.undo_stack.push(self._BeginRemoveColumnCommand(self, column))
        self._step_names_model.removeRow(column)
        self._step_durations_model.removeRow(column)
        for lane_model in self._lane_models:
            lane_model.removeRow(column)
        self.undo_stack.push(self._EndRemoveColumnCommand(self, column))
        self.undo_stack.endMacro()
        return True

    @attrs.frozen(slots=False)
    class _BeginRemoveColumnCommand(QUndoCommand):
        model: TimeLanesModel
        column: int

        def __attrs_post_init__(self):
            super().__init__(f"begin remove column {self.column}")

        def redo(self) -> None:
            assert not self.model._read_only
            self.model.beginRemoveColumns(QModelIndex(), self.column, self.column)

        def undo(self) -> None:
            assert not self.model._read_only
            self.model.endInsertColumns()

    @attrs.frozen(slots=False)
    class _EndRemoveColumnCommand(QUndoCommand):
        model: TimeLanesModel
        column: int

        def __attrs_post_init__(self):
            super().__init__(f"end remove column {self.column}")

        def redo(self) -> None:
            assert not self.model._read_only
            self.model.endRemoveColumns()

        def undo(self) -> None:
            assert not self.model._read_only
            self.model.beginInsertColumns(QModelIndex(), self.column, self.column)

    def lane_number(self) -> int:
        return len(self._lane_models)

    def lane_names(self) -> set[str]:
        """Return the names of the lanes contained in the model."""

        return set(
            model.headerData(0, Qt.Orientation.Horizontal)
            for model in self._lane_models
        )

    def insert_time_lane(
        self, name: str, time_lane: TimeLane, index: Optional[int] = None
    ) -> bool:
        if self._read_only:
            return False
        if index is None:
            index = len(self._lane_models)
        if not (0 <= index <= len(self._lane_models)):
            raise IndexError(f"Index {index} is out of range")
        if len(time_lane) != self.columnCount():
            raise ValueError(
                f"Length of time lane ({len(time_lane)}) does not match "
                f"number of columns ({self.columnCount()})"
            )

        if name in self.lane_names():
            raise ValueError(f"Name {name} is already used")
        lane_model = self._create_lane_model(name, time_lane)
        self.undo_stack.push(self._InsertTimeLaneCommand(self, index, lane_model))
        return True

    @attrs.define(slots=False)
    class _InsertTimeLaneCommand(QUndoCommand):
        model: TimeLanesModel
        index: int
        time_lane_model: TimeLaneModel

        def __attrs_post_init__(self):
            super().__init__(f"insert lane {self.time_lane_model.name()}")

        def redo(self):
            self.model._insert_lane_model(self.index, self.time_lane_model)

        def undo(self):
            self.model._pop_lane_model(self.index)

    def _insert_lane_model(self, index: int, lane_model: TimeLaneModel) -> None:
        assert 0 <= index <= self.lane_number()
        self.beginInsertRows(QModelIndex(), index + 2, index + 2)
        self._lane_models.insert(index, lane_model)
        self.endInsertRows()

    def remove_lane(self, lane_index: int) -> bool:
        """Remove a lane from the model.

        This method replaces :meth:`removeRow` as not all rows can be removed.
        Push the action on the model undo stack.
        """

        if self._read_only:
            return False
        if not (0 <= lane_index < self.lane_number()):
            return False
        self.undo_stack.push(self._RemoveTimeLaneCommand(self, lane_index))
        return True

    class _RemoveTimeLaneCommand(QUndoCommand):
        def __init__(self, model: "TimeLanesModel", index: int):
            self.name = model.get_lane_name(index)
            super().__init__(f"remove lane {self.name}")
            self.model = model
            self.index = index
            self.time_lane_model: TimeLaneModel | None = None

        def redo(self):
            self.time_lane_model = self.model._pop_lane_model(self.index)

        def undo(self):
            assert self.time_lane_model is not None
            self.model._insert_lane_model(self.index, self.time_lane_model)
            self.time_lane_model = None

    def _pop_lane_model(self, lane_index: int) -> TimeLaneModel:
        assert 0 <= lane_index < len(self._lane_models)
        self.beginRemoveRows(QModelIndex(), lane_index + 2, lane_index + 2)
        model = self._lane_models.pop(lane_index)
        self.endRemoveRows()
        return model

    def get_cell_context_actions(self, index: QModelIndex) -> list[QAction | QMenu]:
        if not index.isValid():
            return []
        if self._read_only:
            return []
        if index.row() >= 2:
            return self._lane_models[index.row() - 2].get_cell_context_actions(
                self._map_to_source(index)
            )
        else:
            return []

    def get_lane_header_context_actions(self, lane_index: int) -> list[QAction | QMenu]:
        if not 0 <= lane_index < len(self._lane_models):
            return []
        if self._read_only:
            return []
        return self._lane_models[lane_index].get_header_context_actions()

    def span(self, index):
        if not index.isValid():
            return QSize(1, 1)
        if index.row() >= 2:
            mapped_index = self._map_to_source(index)
            span = self._lane_models[index.row() - 2].span(mapped_index)
            return QSize(span.height(), span.width())
        return QSize(1, 1)

    def expand_step(self, step: int, lane_index: int, start: int, stop: int) -> bool:
        if self._read_only:
            return False
        if not (0 <= step < self.columnCount()):
            return False
        if not (0 <= lane_index < self.lane_number()):
            return False
        lane_model = self._lane_models[lane_index]
        return lane_model.expand_step(step, start, stop)

    def _map_to_source(self, index: QModelIndex | QPersistentModelIndex) -> QModelIndex:
        assert index.isValid()
        assert self.hasIndex(index.row(), index.column())
        if index.row() == 0:
            return self._step_names_model.index(index.column(), 0)
        elif index.row() == 1:
            return self._step_durations_model.index(index.column(), 0)
        else:
            return self._lane_models[index.row() - 2].index(index.column(), 0)

    def simplify(self) -> bool:
        if self._read_only:
            return False
        self.undo_stack.beginMacro("simplify")
        for lane_model in self._lane_models:
            lane_model.simplify()
        self.undo_stack.endMacro()
        return True


def get_lane_model_name(model: TimeLaneModel) -> str:
    lane_name = model.headerData(
        0, Qt.Orientation.Horizontal, Qt.ItemDataRole.DisplayRole
    )
    if not isinstance(lane_name, str):
        raise TypeError(f"Expected str, got {type(lane_name)}")
    return lane_name



from typing import assert_never, TypedDict

import numpy as np

from caqtus.device import DeviceName
from caqtus.shot_compilation import (
    SequenceContext,
    DeviceNotUsedException,
    ShotContext,
)
from caqtus.shot_compilation.timed_instructions import (
    TimedInstruction,
    Pattern,
    concatenate,
)
from caqtus.types.recoverable_exceptions import InvalidValueError
from caqtus.types.timelane import CameraTimeLane, TakePicture
from ._configuration import CameraConfiguration
from ..sequencer import TimeStep
from ..sequencer.compilation import TriggerableDeviceCompiler
from ..sequencer.timing import ns, number_time_steps_between
from ...types.image.roi import RectangularROI


class CameraCompiler(TriggerableDeviceCompiler):
    """Computes parameters for a camera device."""

    def __init__(self, device_name: DeviceName, sequence_context: SequenceContext):
        super().__init__(device_name, sequence_context)
        self.__device_name = device_name
        try:
            lane = sequence_context.get_lane(device_name)
        except KeyError:
            raise DeviceNotUsedException(device_name) from None
        if not isinstance(lane, CameraTimeLane):
            raise TypeError(
                f"Expected a camera time lane for device {device_name}, got "
                f"{type(lane)}"
            )
        self.__lane = lane
        configuration = sequence_context.get_device_configuration(device_name)
        if not isinstance(configuration, CameraConfiguration):
            raise TypeError(
                f"Expected a camera configuration for device {device_name}, got "
                f"{type(configuration)}"
            )
        self.__configuration = configuration

    class CameraInitializationParameters(TypedDict):
        """The parameters to pass to the camera constructor.

        This dictionary contains the following keys:

            * roi: The region of interest of the sensor for the camera during the
                sequence.
            * external_trigger: Whether the camera should be triggered externally or
                not.
            * timeout: The maximum time to wait for the camera to be ready to take a
                picture.
        """

        roi: RectangularROI
        external_trigger: bool
        timeout: float

    def compile_initialization_parameters(self) -> CameraInitializationParameters:
        """Compile the parameters to pass to the device constructor."""

        return self.CameraInitializationParameters(
            roi=self.__configuration.roi,
            external_trigger=True,
            timeout=1.0,
        )

    class CameraShotParameters(TypedDict):
        """The parameters to pass to the camera controller for a shot.

        This dictionary contains the following keys:

            * timeout: The maximum time to wait for a picture to be taken.
            * picture_names: The names of the pictures to take, in the order they
                should be taken.
            * exposures: The exposure times of the pictures to take, in the order they
                should be taken.
                The number of exposures matches the number of pictures.
        """

        timeout: float
        picture_names: list[str]
        exposures: list[float]

    def compile_shot_parameters(
        self, shot_context: ShotContext
    ) -> CameraShotParameters:
        """Compile the parameters to pass to the camera controller for a shot.

        The exposures for the pictures are computed from the duration of the
        corresponding blocks in the camera time lane.
        """

        step_durations = shot_context.get_step_durations()
        exposures: list[float] = []
        picture_names = []
        shot_context.mark_lane_used(self.__device_name)
        for value, (start, stop) in zip(
            self.__lane.block_values(), self.__lane.block_bounds(), strict=True
        ):
            if isinstance(value, TakePicture):
                exposure = sum(step_durations[start:stop])
                exposures.append(float(exposure))
                picture_names.append(value.picture_name)
        return self.CameraShotParameters(
            # Add a bit of extra time to the timeout, in case the shot takes a bit of
            # time to actually start.
            timeout=float(shot_context.get_shot_duration() + 1),
            picture_names=picture_names,
            exposures=exposures,
        )

    def compute_trigger(
        self, sequencer_time_step: TimeStep, shot_context: ShotContext
    ) -> TimedInstruction[np.bool_]:
        """Compute the trigger for the camera.

        For a camera, the trigger is high during the exposure time of each picture,
        and low in between.
        """

        step_bounds = shot_context.get_step_start_times()

        instructions: list[TimedInstruction[np.bool_]] = []
        for value, (start, stop) in zip(
            self.__lane.block_values(), self.__lane.block_bounds(), strict=True
        ):
            length = number_time_steps_between(
                step_bounds[start], step_bounds[stop], sequencer_time_step
            )
            if isinstance(value, TakePicture):
                if length == 0:
                    raise InvalidValueError(
                        f"No trigger can be generated for picture "
                        f"'{value.picture_name}' because its exposure is too short"
                        f"({(step_bounds[stop] - step_bounds[start]) / ns} ns) with "
                        f"respect to the time step ({sequencer_time_step} ns)"
                    )
                instructions.append(Pattern([True]) * length)
            elif value is None:
                instructions.append(Pattern([False]) * length)
            else:
                assert_never(value)
        return concatenate(*instructions)



import functools
from typing import Optional

from PySide6 import QtCore
from PySide6.QtCore import QSortFilterProxyModel, QModelIndex
from PySide6.QtGui import QAction
from PySide6.QtWidgets import QTreeView, QMenu, QWidget

from caqtus.session import ExperimentSessionMaker, PureSequencePath, PathNotFoundError
from caqtus.utils.result import is_failure_type
from .async_model import AsyncPathHierarchyModel
from .delegate import ProgressDelegate


class AsyncPathHierarchyView(QTreeView):
    sequence_double_clicked = QtCore.Signal(PureSequencePath)

    def __init__(
        self, session_maker: ExperimentSessionMaker, parent: Optional[QWidget] = None
    ):
        super().__init__(parent)
        self.session_maker = session_maker
        self._model = AsyncPathHierarchyModel(session_maker, self)
        self._proxy_model = QSortFilterProxyModel(self)
        self._proxy_model.setSourceModel(self._model)
        self.setModel(self._proxy_model)
        self.setSortingEnabled(True)
        self.header().setContextMenuPolicy(
            QtCore.Qt.ContextMenuPolicy.CustomContextMenu
        )
        self.header().customContextMenuRequested.connect(self.show_header_menu)
        self.doubleClicked.connect(self.on_double_click)
        self.sortByColumn(4, QtCore.Qt.SortOrder.AscendingOrder)
        self.hideColumn(4)
        self.setItemDelegateForColumn(1, ProgressDelegate(self))
        self.setUniformRowHeights(True)

    async def run_async(self) -> None:
        await self._model.run()

    def show_header_menu(self, pos):
        menu = QMenu(self)
        visibility_menu = menu.addMenu("Visible")
        # The first column is the name and should not be hidden.
        for column in range(1, self.model().columnCount()):
            action = QAction(
                self.model().headerData(column, QtCore.Qt.Orientation.Horizontal), self
            )
            action.setCheckable(True)
            action.setChecked(not self.isColumnHidden(column))
            action.triggered.connect(functools.partial(self.toggle_visibility, column))
            visibility_menu.addAction(action)
        menu.exec(self.mapToGlobal(pos))

    def toggle_visibility(self, column: int):
        column_hidden = self.isColumnHidden(column)
        self.setColumnHidden(column, not column_hidden)

    def on_double_click(self, index: QModelIndex):
        path = self._model.get_path(self._proxy_model.mapToSource(index))
        with self.session_maker() as session:
            is_sequence_result = session.sequences.is_sequence(path)
            if is_failure_type(is_sequence_result, PathNotFoundError):
                # It can happen that the path is not found, if it was deleted in the
                # background, but the user double-clicked it before the model was
                # updated.
                # In this case, we just ignore the double click.
                return
            is_sequence = is_sequence_result.value
        if is_sequence:
            self.sequence_double_clicked.emit(path)



from collections.abc import Mapping

from ._units import UNITS, Unit, ureg

# TODO: Some tests fail when trying to replace `getattr(ureg, unit)` with `Unit(unit)`.
#  To check.
units: Mapping[str, Unit] = {unit: getattr(ureg, unit) for unit in UNITS}



"""This module contains the :class:`DeviceConfiguration` class."""

from __future__ import annotations

from typing import (
    TypeVar,
    Optional,
    NewType,
    Generic,
)

import attrs

from caqtus.device.runtime import Device

DeviceServerName = NewType("DeviceServerName", str)

DeviceType = TypeVar("DeviceType", bound=Device)


@attrs.define
class DeviceConfiguration(Generic[DeviceType]):
    """Contains static information about a device.

    This is an abstract class, generic in :data:`DeviceType` that stores the information
    necessary to connect to a device and program it during a sequence.

    This information is meant to be encoded in a user-friendly way that might not be
    possible to be directly programmed on a device.
    For example, it might contain not yet evaluated
    :class:`caqtus.types.expression.Expression` objects that only make sense in the
    context of a shot.

    Subclasses should add necessary attributes depending on the device.

    The dunder method :meth:`__eq__` should be implemented.

    Attributes:
        remote_server: Indicates the name of the computer on which the device should be
            instantiated.
    """

    remote_server: Optional[DeviceServerName] = attrs.field()


DeviceConfigType = TypeVar("DeviceConfigType", bound=DeviceConfiguration)



from __future__ import annotations

import functools
from collections.abc import Iterable, Callable, Generator
from collections.abc import Mapping, Iterator
from typing import TypeAlias, TypeGuard, Any, assert_type, override, assert_never, Self

import attrs
import numpy

import caqtus.formatter as fmt
from caqtus.types.expression import Expression
from caqtus.types.parameter import (
    NotAnalogValueError,
    Parameter,
    ParameterSchema,
)
from caqtus.types.parameter import is_parameter
from caqtus.types.recoverable_exceptions import InvalidTypeError
from caqtus.utils import serialization
from ._step_context import StepContext
from .iteration_configuration import IterationConfiguration, Unknown
from ..parameter._analog_value import is_scalar_analog_value, ScalarAnalogValue
from ..recoverable_exceptions import EvaluationError
from ..units import (
    DimensionalityError,
    InvalidDimensionalityError,
    dimensionless,
    Quantity,
    Unit,
)
from ..variable_name import DottedVariableName


def validate_step(instance, attribute, step):
    if is_step(step):
        return
    else:
        raise TypeError(f"Invalid step: {step}")


@attrs.define
class ContainsSubSteps:
    sub_steps: list[Step] = attrs.field(
        validator=attrs.validators.deep_iterable(
            iterable_validator=attrs.validators.instance_of(list),
            member_validator=validate_step,
        ),
        on_setattr=attrs.setters.validate,
    )


@attrs.define
class VariableDeclaration:
    """Represents the declaration of a variable.

    Attributes:
        variable: The name of the variable.
        value: The unevaluated to assign to the variable.
    """

    __match_args__ = ("variable", "value")

    variable: DottedVariableName = attrs.field(
        validator=attrs.validators.instance_of(DottedVariableName),
        on_setattr=attrs.setters.validate,
    )
    value: Expression = attrs.field(
        validator=attrs.validators.instance_of(Expression),
        on_setattr=attrs.setters.validate,
    )

    def __str__(self):
        return f"{self.variable} = {self.value}"


@attrs.define
class LinspaceLoop(ContainsSubSteps):
    """Represents a loop that iterates between two values with a fixed number of steps.

    Attributes:
        variable: The name of the variable that is being iterated over.
        start: The start value of the variable.
        stop: The stop value of the variable.
        num: The number of steps to take between the start and stop values.
    """

    __match_args__ = (
        "variable",  # pyright: ignore[reportAssignmentType]
        "start",
        "stop",
        "num",
        "sub_steps",
    )
    variable: DottedVariableName = attrs.field(
        validator=attrs.validators.instance_of(DottedVariableName),
        on_setattr=attrs.setters.validate,
    )
    start: Expression = attrs.field(
        validator=attrs.validators.instance_of(Expression),
        on_setattr=attrs.setters.validate,
    )
    stop: Expression = attrs.field(
        validator=attrs.validators.instance_of(Expression),
        on_setattr=attrs.setters.validate,
    )
    num: int = attrs.field(
        converter=int,
        validator=attrs.validators.ge(0),
        on_setattr=attrs.setters.pipe(attrs.setters.convert, attrs.setters.validate),
    )

    def __str__(self):
        return f"linspace loop over {self.variable}"

    def loop_values(
        self, evaluation_context: Mapping[DottedVariableName, Any]
    ) -> Iterator[ScalarAnalogValue]:
        """Returns the values that the variable represented by this loop takes.

        Args:
            evaluation_context: Contains the value of the variables with which to
                evaluate the start and stop expressions of the loop.

        Raises:
            EvaluationError: if the start or stop expressions could not be evaluated.
            NotAnalogValueError: if the start or stop expressions don't evaluate to an
                analog value.
            DimensionalityError: if the start or stop values are not commensurate.
        """

        try:
            start = _to_scalar_analog_value(self.start.evaluate(evaluation_context))
        except NotAnalogValueError:
            raise NotAnalogValueError(
                f"Start {fmt.expression(self.start)} of {self} does not evaluate to an "
                f"analog value"
            ) from None
        try:
            stop = _to_scalar_analog_value(self.stop.evaluate(evaluation_context))
        except NotAnalogValueError:
            raise NotAnalogValueError(
                f"Stop {fmt.expression(self.stop)} of {self} does not evaluate to an "
                f"analog value"
            ) from None

        # Here we enforce that the values generated have the same format as the start
        # value.
        if isinstance(start, float):
            try:
                stop = _to_dimensionless_float(stop)
            except DimensionalityError:
                raise InvalidDimensionalityError(
                    f"Start {fmt.expression(self.start)} of {self} is "
                    f"dimensionless, but stop {fmt.expression(self.stop)} cannot "
                    f"be converted to dimensionless"
                ) from None
            assert_type(stop, float)
            assert_type(start, float)
            for value in numpy.linspace(start, stop, self.num):
                yield float(value.item())
        elif isinstance(start, int):
            raise AssertionError("start must be strictly a float or a Quantity")
        else:
            try:
                stop = _to_unit(stop, start.units)
            except DimensionalityError as e:
                raise InvalidDimensionalityError(
                    f"Start {fmt.expression(self.start)} of {self} has invalid "
                    f"dimensionality."
                ) from e
            assert_type(start, Quantity[float])
            assert_type(stop, Quantity[float])
            for value in numpy.linspace(start.magnitude, stop.magnitude, self.num):
                yield Quantity(float(value), start.units)


@attrs.define
class ArangeLoop(ContainsSubSteps):
    """Represents a loop that iterates between two values with a fixed step size.

    Attributes:
        variable: The name of the variable that is being iterated over.
        start: The start value of the variable.
        stop: The stop value of the variable.
        step: The step size between each value.
    """

    __match_args__ = (
        "variable",  # pyright: ignore[reportAssignmentType]
        "start",
        "stop",
        "step",
        "sub_steps",
    )
    variable: DottedVariableName = attrs.field(
        validator=attrs.validators.instance_of(DottedVariableName),
        on_setattr=attrs.setters.validate,
    )
    start: Expression = attrs.field(
        validator=attrs.validators.instance_of(Expression),
        on_setattr=attrs.setters.validate,
    )
    stop: Expression = attrs.field(
        validator=attrs.validators.instance_of(Expression),
        on_setattr=attrs.setters.validate,
    )
    step: Expression = attrs.field(
        validator=attrs.validators.instance_of(Expression),
        on_setattr=attrs.setters.validate,
    )

    def __str__(self):
        return f"arange loop over {fmt.shot_param(self.variable)}"

    def loop_values(
        self, evaluation_context: Mapping[DottedVariableName, Any]
    ) -> Iterator[ScalarAnalogValue]:
        """Returns the values that the variable represented by this loop takes.

        Args:
            evaluation_context: Contains the value of the variables with which to
                evaluate the start, stop and step expressions of the loop.

        Raises:
            EvaluationError: if the start, stop or step expressions could not be
                evaluated.
            NotAnalogValueError: if the start, stop or step expressions don't evaluate
                to an analog value.
            InvalidDimensionalityError: if the start, stop and step values are not
                commensurate.
        """

        try:
            start = _to_scalar_analog_value(self.start.evaluate(evaluation_context))
        except NotAnalogValueError:
            raise NotAnalogValueError(
                f"Start {fmt.expression(self.start)} of {self} does not evaluate to an "
                f"analog value"
            ) from None
        try:
            stop = _to_scalar_analog_value(self.stop.evaluate(evaluation_context))
        except NotAnalogValueError:
            raise NotAnalogValueError(
                f"Stop {fmt.expression(self.stop)} of {self} does not evaluate to an "
                f"analog value"
            ) from None
        try:
            step = _to_scalar_analog_value(self.step.evaluate(evaluation_context))
        except NotAnalogValueError:
            raise NotAnalogValueError(
                f"Step {fmt.expression(self.step)} of {self} does not evaluate to an "
                f"analog value"
            ) from None

        # Here we enforce that the values generated have the same format as the start
        # value.
        if isinstance(start, float):
            try:
                stop = _to_dimensionless_float(stop)
            except DimensionalityError:
                raise InvalidDimensionalityError(
                    f"Start {fmt.expression(self.start)} of {self} is "
                    f"dimensionless, but stop {fmt.expression(self.stop)} cannot "
                    f"be converted to dimensionless"
                ) from None
            try:
                step = _to_dimensionless_float(step)
            except DimensionalityError:
                raise InvalidDimensionalityError(
                    f"Step {fmt.expression(self.step)} of {self} is "
                    f"dimensionless, but stop {fmt.expression(self.stop)} cannot "
                    f"be converted to dimensionless"
                ) from None
            assert_type(start, float)
            assert_type(stop, float)
            assert_type(step, float)
            for value in numpy.arange(start, stop, step):
                yield float(value)
        elif isinstance(start, int):
            raise AssertionError("start must be strictly a float or a Quantity")
        else:
            try:
                stop = _to_unit(stop, start.units)
            except DimensionalityError as e:
                raise InvalidDimensionalityError(
                    f"Start {fmt.expression(self.start)} of {self} has invalid "
                    f"dimensionality."
                ) from e
            try:
                step = _to_unit(step, start.units)
            except DimensionalityError as e:
                raise InvalidDimensionalityError(
                    f"Step {fmt.expression(self.step)} of {self} has invalid "
                    f"dimensionality."
                ) from e
            assert_type(start, Quantity[float])
            assert_type(stop, Quantity[float])
            assert_type(step, Quantity[float])
            for value in numpy.arange(start.magnitude, stop.magnitude, step.magnitude):
                yield Quantity(float(value), start.units)


@attrs.define
class ExecuteShot:
    """Step that represents the execution of a shot."""

    def __str__(self):
        return "do shot"


def unstructure_hook(execute_shot: ExecuteShot):
    return {"execute": "shot"}


def structure_hook(data: str, cls: type[ExecuteShot]) -> ExecuteShot:
    return ExecuteShot()


serialization.register_unstructure_hook(ExecuteShot, unstructure_hook)

serialization.register_structure_hook(ExecuteShot, structure_hook)

"""TypeAlias for the different types of steps."""
Step: TypeAlias = ExecuteShot | VariableDeclaration | LinspaceLoop | ArangeLoop


def is_step(step) -> TypeGuard[Step]:
    return isinstance(
        step,
        (
            ExecuteShot,
            VariableDeclaration,
            LinspaceLoop,
            ArangeLoop,
        ),
    )


@attrs.define
class StepsConfiguration(IterationConfiguration):
    """Define the parameter iteration of a sequence as a list of steps.

    Attributes:
        steps: The steps of the iteration.
    """

    steps: list[Step] = attrs.field(
        validator=attrs.validators.deep_iterable(
            iterable_validator=attrs.validators.instance_of(list),
            member_validator=validate_step,
        ),
        on_setattr=attrs.setters.validate,
    )

    @classmethod
    def empty(cls) -> Self:
        return cls(steps=[])

    def expected_number_shots(self) -> int | Unknown:
        """Returns the expected number of shots that will be executed by the sequence.

        Returns:
            A positive integer if the number of shots can be determined, or Unknown if
            the number of shots cannot be determined.
        """

        return sum(expected_number_shots(step) for step in self.steps)

    def get_parameter_names(self) -> set[DottedVariableName]:
        return set().union(*[get_parameter_names(step) for step in self.steps])

    @classmethod
    def dump(cls, steps_configuration: StepsConfiguration) -> serialization.JSON:
        return serialization.unstructure(steps_configuration, StepsConfiguration)

    @classmethod
    def load(cls, data: serialization.JSON) -> StepsConfiguration:
        return serialization.structure(data, StepsConfiguration)

    def walk(self, initial_context: StepContext) -> Iterator[StepContext]:
        """Returns the context for every shot encountered while walking the steps."""

        return walk_steps(self.steps, initial_context)

    @override
    def get_parameter_schema(
        self, initial_parameters: Mapping[DottedVariableName, Parameter]
    ) -> ParameterSchema:
        context_iterator = self.walk(StepContext(initial_parameters))
        try:
            first_context = next(context_iterator)
        except StopIteration:
            # In case there is not steps to walk, we return a schema made only of the
            # initial constant parameters.
            return ParameterSchema(
                _constant_schema=initial_parameters, _variable_schema={}
            )
        variable_parameters = self.get_parameter_names()
        constant_parameters = set(initial_parameters) - variable_parameters
        constant_schema = {
            name: first_context.variables[name] for name in constant_parameters
        }
        initial_values = first_context.variables.to_flat_dict()
        variable_schema = {
            name: ParameterSchema.type_from_value(initial_values[name])
            for name in variable_parameters
        }
        return ParameterSchema(
            _constant_schema=constant_schema, _variable_schema=variable_schema
        )


@functools.singledispatch
def expected_number_shots(step: Step) -> int | Unknown:
    raise NotImplementedError(f"Cannot determine the number of shots for {step}")


@expected_number_shots.register
def _(step: VariableDeclaration):
    return 0


@expected_number_shots.register
def _(step: ExecuteShot):
    return 1


@expected_number_shots.register
def _(step: LinspaceLoop):
    sub_steps_number = sum(
        expected_number_shots(sub_step) for sub_step in step.sub_steps
    )
    return sub_steps_number * step.num


@expected_number_shots.register
def _(step: ArangeLoop):
    try:
        length = len(list(step.loop_values({})))
    except (EvaluationError, NotAnalogValueError, InvalidDimensionalityError):
        # The errors above can occur if the steps are still being edited or if the
        # expressions depend on other variables that are not defined here.
        # These can be errors on the user side, so we don't want to crash on them, and
        # we just indicate that we don't know the number of shots.
        return Unknown()

    sub_steps_number = sum(
        expected_number_shots(sub_step) for sub_step in step.sub_steps
    )
    return sub_steps_number * length


def get_parameter_names(step: Step) -> set[DottedVariableName]:
    match step:
        case VariableDeclaration(variable=variable, value=_):
            return {variable}
        case ExecuteShot():
            return set()
        case LinspaceLoop(variable=variable, sub_steps=sub_steps) | ArangeLoop(
            variable=variable, sub_steps=sub_steps
        ):
            return {variable}.union(
                *[get_parameter_names(sub_step) for sub_step in sub_steps]
            )
        case _:
            assert_never(step)


def wrap_error[
    S: Step
](
    function: Callable[[S, StepContext], Generator[StepContext, None, StepContext]]
) -> Callable[[S, StepContext], Generator[StepContext, None, StepContext]]:
    """Wrap a function that evaluates a step to raise nicer errors for the user."""

    @functools.wraps(function)
    def wrapper(step: S, context: StepContext):
        try:
            return function(step, context)
        except Exception as e:
            raise StepEvaluationError(f"Error while evaluating step <{step}>") from e

    return wrapper


def walk_steps(
    steps: Iterable[Step], initial_context: StepContext
) -> Iterator[StepContext]:
    """Yields the context for each shot defined by the steps.

    This function will recursively evaluate each step in the sequence passed as
    argument.
    Before executing the sequence, an empty context is initialized.
    The context holds the value of the parameters at a given point in the sequence.
    Each step has the possibility to update the context with new values.
    """

    context = initial_context

    for step in steps:
        context = yield from walk_step(step, context)


@functools.singledispatch
@wrap_error
def walk_step(
    step: Step, context: StepContext
) -> Generator[StepContext, None, StepContext]:
    """Iterates over the steps of a sequence.

    Args:
        step: the step of the sequence currently executed
        context: Contains the values of the variables before this step.

    Yields:
        The context for every shot encountered while walking the steps.

    Returns:
        The context after the step passed in argument has been executed.
    """

    raise NotImplementedError(f"Cannot walk step {step}")


# noinspection PyUnreachableCode
@walk_step.register
@wrap_error
def _(
    declaration: VariableDeclaration,
    context: StepContext,
) -> Generator[StepContext, None, StepContext]:
    """Execute a VariableDeclaration step.

    This step updates the context passed with the value of the variable declared.
    """

    value = declaration.value.evaluate(context.variables.dict())
    if not is_parameter(value):
        raise InvalidTypeError(
            f"{fmt.expression(declaration.value)}> does not evaluate to a parameter, "
            f"but to {fmt.type_(type(value))}.",
        )
    return context.update_variable(declaration.variable, value)

    # This code is unreachable, but it is kept here to make the function a generator.
    if False:
        yield context


@walk_step.register
@wrap_error
def _(
    arange_loop: ArangeLoop,
    context: StepContext,
) -> Generator[StepContext, None, StepContext]:
    """Loop over a variable in a numpy arange like loop.

    This function will loop over the variable defined in the arange loop and execute
    the sub steps for each value of the variable.

    Returns:
        A new context object containing the value of the arange loop variable after
        the last iteration, plus the values of the variables defined in the sub
        steps.
    """

    for value in arange_loop.loop_values(context.variables.dict()):
        context = context.update_variable(arange_loop.variable, value)
        for step in arange_loop.sub_steps:
            context = yield from walk_step(step, context)
    return context


@walk_step.register
@wrap_error
def _(
    linspace_loop: LinspaceLoop,
    context: StepContext,
) -> Generator[StepContext, None, StepContext]:
    """Loop over a variable in a numpy linspace like loop.

    This function will loop over the variable defined in the linspace loop and
    execute the sub steps for each value of the variable.

    Returns:
        A new context object containing the value of the linspace loop variable
        after the last iteration, plus the values of the variables defined in the
        sub steps.
    """

    for value in linspace_loop.loop_values(context.variables.dict()):
        context = context.update_variable(linspace_loop.variable, value)
        for step in linspace_loop.sub_steps:
            context = yield from walk_step(step, context)
    return context


@walk_step.register
@wrap_error
def _(
    shot: ExecuteShot, context: StepContext
) -> Generator[StepContext, None, StepContext]:
    """Schedule a shot to be run.

    This function schedule to run a shot on the experiment with the parameters
    defined in the context at this point.

    Returns:
        The context passed as argument unchanged.
    """

    yield context
    return context


class StepEvaluationError(Exception):
    pass


def _to_scalar_analog_value(value: Any) -> ScalarAnalogValue:
    """Attempt to convert a value to a scalar analog value.

    Raises:
        NotAnalogValueError: If the value can't be converted to a scalar analog value.
    """

    if isinstance(value, int):
        return float(value)
    if not is_scalar_analog_value(value):
        raise NotAnalogValueError(value)
    return value


def _to_dimensionless_float(value: ScalarAnalogValue) -> float:
    """Convert a scalar analog value to a dimensionless float.

    Raises:
        DimensionalityError: If the value is a quantity not commensurate with
        dimensionless.
    """

    if isinstance(value, Quantity):
        return value.to_unit(dimensionless).magnitude
    elif isinstance(value, float):
        return value
    elif isinstance(value, int):
        raise AssertionError("stop must be strictly a float or a Quantity")
    else:
        assert_never(value)


def _to_unit[U: Unit](value: ScalarAnalogValue, unit: U) -> Quantity[float, U]:
    """Convert a scalar analog value to the same unit as another.

    Raises:
        DimensionalityError: If the value is a quantity not commensurate with the unit.
    """

    if isinstance(value, float):
        value = Quantity(value, dimensionless)
    elif isinstance(value, int):
        raise AssertionError("stop must be strictly a float or a Quantity")
    return value.to_unit(unit)



from .configuration import (
    SequencerConfiguration,
    ChannelConfiguration,
    AnalogChannelConfiguration,
    DigitalChannelConfiguration,
)


__all__ = [
    "SequencerConfiguration",
    "ChannelConfiguration",
    "AnalogChannelConfiguration",
    "DigitalChannelConfiguration",
]



from __future__ import annotations

import itertools
from typing import Optional

import yaml
from PySide6.QtCore import Qt, QModelIndex, QRect
from PySide6.QtGui import QAction, QFont, QUndoStack, QCursor
from PySide6.QtWidgets import (
    QTableView,
    QMenu,
    QWidget,
    QVBoxLayout,
    QToolBar,
    QDialog,
    QMessageBox,
    QApplication,
)

from caqtus.device import DeviceConfiguration, DeviceName
from caqtus.gui.condetrol._icons import get_icon
from caqtus.gui.qtutil import block_signals, temporary_widget
from caqtus.types.timelane import TimeLanes, TimeLane
from ._delegate import TimeLaneDelegate
from ._time_lanes_model import TimeLanesModel
from .add_lane_dialog import AddLaneDialog
from .extension import CondetrolLaneExtensionProtocol


class TimeLanesEditor(QWidget):
    """A widget for editing the time lanes of a sequence.

    Signals:
        time_lanes_edited: Emitted when the user edits the time lanes.
    """

    def __init__(
        self,
        extension: CondetrolLaneExtensionProtocol,
        device_configurations: dict[DeviceName, DeviceConfiguration],
        parent: Optional[QWidget] = None,
    ):
        super().__init__(parent)
        self.view = TimeLanesView(
            extension=extension,
            device_configurations=device_configurations,
            parent=self,
        )
        self._extension = extension
        self.toolbar = QToolBar(self)
        self.toolbar.setFloatable(False)
        self.toolbar.setMovable(False)
        self.add_lane_action = self.toolbar.addAction(
            get_icon("add-time-lane", self.palette().buttonText().color()), "Add lane"
        )
        self.add_lane_action.triggered.connect(self._on_add_lane_triggered)
        self.simplify_action = self.toolbar.addAction(
            get_icon("simplify-timelanes", self.palette().buttonText().color()),
            "Simplify",
        )
        self.simplify_action.triggered.connect(self._simplify_timelanes)
        self.toolbar.addSeparator()
        self.copy_to_clipboard_action = self.toolbar.addAction(
            get_icon("copy", self.palette().buttonText().color()), "Copy to clipboard"
        )
        self.copy_to_clipboard_action.triggered.connect(self.copy_to_clipboard)
        self.paste_from_clipboard_action = self.toolbar.addAction(
            get_icon("paste", self.palette().buttonText().color()),
            "Paste from clipboard",
        )
        self.paste_from_clipboard_action.triggered.connect(self.paste_from_clipboard)
        layout = QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(0)
        layout.addWidget(self.view)
        self.setLayout(layout)

        self._add_lane_dialog = AddLaneDialog(self)
        self._add_lane_dialog.set_lane_types(extension.available_new_lanes())

        font = QFont("JetBrains Mono")
        self.setFont(font)

    @property
    def undo_stack(self) -> QUndoStack:
        return self.view.undo_stack

    def set_read_only(self, read_only: bool) -> None:
        """Set the editor to read-only mode.

        In read-only mode, the user cannot edit the time lanes.
        """

        self.view.set_read_only(read_only)

        self.add_lane_action.setEnabled(not read_only)
        self.simplify_action.setEnabled(not read_only)
        self.paste_from_clipboard_action.setEnabled(not read_only)

    def model(self) -> TimeLanesModel:
        return self.view.model()

    def set_time_lanes(self, time_lanes: TimeLanes) -> None:
        """Set the time lanes to be edited.

        The signal time_lanes_edited is not emitted when this method is called.
        """

        with block_signals(self):
            self.model().set_timelanes(time_lanes)

    def get_time_lanes(self) -> TimeLanes:
        """Return a copy of the time lanes currently being edited."""

        return self.model().get_timelanes()

    def has_uncommitted_edits(self) -> bool:
        return self.model().has_uncommitted_edits()

    def commit_edits(self) -> None:
        self.model().commit_edits()

    def _simplify_timelanes(self):
        self.model().simplify()

    def _on_add_lane_triggered(self) -> None:
        if self._add_lane_dialog.exec() == QDialog.DialogCode.Accepted:
            lane_name = self._add_lane_dialog.get_lane_name()
            lane_type = self._add_lane_dialog.get_lane_type()
            if not lane_name:
                return
            if lane_name in self.view.get_time_lanes().lanes:
                QMessageBox.warning(  # type: ignore[reportCallIssue]
                    self,
                    "Lane already exists",
                    f"Can't add the lane <i>{lane_name}</i> because there is already "
                    "a lane with this name.",
                )
            else:
                lane = self._extension.create_new_lane(
                    lane_type, self.view.model().columnCount()
                )
                self.view.add_lane(lane_name, lane)

    def copy_to_clipboard(self):
        time_lanes = self.view.get_time_lanes()
        unstructured = self._extension.unstructure_time_lanes(time_lanes)

        text = yaml.safe_dump(unstructured)
        clipboard = QApplication.clipboard()
        clipboard.setText(text)

    def paste_from_clipboard(self) -> bool:
        """Paste the content of the clipboard into the editor.

        Returns:
            True if the content was successfully pasted, False otherwise.
        """

        if self.view.is_read_only():
            return False

        clipboard = QApplication.clipboard()
        text = clipboard.text()
        try:
            content = yaml.safe_load(text)
        except yaml.YAMLError as e:
            QMessageBox.warning(  # type: ignore[reportCallIssue]
                self,
                "Invalid YAML content",
                f"Could not parse the clipboard content as YAML:\n {e}",
            )
            return False

        # TODO: raise recoverable error if the content is not valid
        time_lanes = self._extension.structure_time_lanes(content)

        self.model().set_timelanes_with_undo(time_lanes, "paste from clipboard")
        return True


class OverlayStepsView(QTableView):
    def __init__(self, model: TimeLanesModel, parent: TimeLanesView):
        super().__init__(parent)
        self.setModel(model)

        model.rowsInserted.connect(self.hide_lanes)
        model.modelReset.connect(self.hide_lanes)
        self.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.setSelectionModel(parent.selectionModel())
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self.setVerticalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self.horizontalHeader().hide()
        self.verticalHeader().setContextMenuPolicy(
            Qt.ContextMenuPolicy.CustomContextMenu
        )

        parent.horizontalHeader().sectionResized.connect(self.resize_columns)
        parent.verticalHeader().sectionResized.connect(self.resize_rows)
        self.verticalHeader().sectionResized.connect(self.resize_parent_rows)
        parent.horizontalScrollBar().valueChanged.connect(
            self.horizontalScrollBar().setValue
        )
        self.horizontalScrollBar().valueChanged.connect(
            parent.horizontalScrollBar().setValue
        )

        for column in range(model.columnCount()):
            self.setColumnWidth(column, parent.columnWidth(column))

        self.setRowHeight(0, parent.rowHeight(0))
        self.setRowHeight(1, parent.rowHeight(1))

        self.update_geometry()

        self.verticalScrollBar().valueChanged.connect(
            lambda _: self.verticalScrollBar().setValue(
                self.verticalScrollBar().minimum()
            )
        )
        self.verticalHeader().customContextMenuRequested.connect(
            self._on_steps_context_menu_requested
        )

    def _on_steps_context_menu_requested(self, _):
        if self.is_read_only():
            return

        with temporary_widget(QMenu(self.horizontalHeader())) as menu:
            add_step_action = QAction("Add step")
            menu.addAction(add_step_action)
            add_step_action.triggered.connect(
                lambda: self.model().insertColumn(
                    self.model().columnCount(), QModelIndex()
                )
            )
            menu.exec(QCursor.pos())

    def model(self) -> TimeLanesModel:
        model = super().model()
        assert isinstance(model, TimeLanesModel)
        return model

    def is_read_only(self) -> bool:
        return self.model().is_read_only()

    def parent(self) -> TimeLanesView:
        p = super().parent()
        assert isinstance(p, TimeLanesView)
        return p

    def resize_columns(self, logical_index: int, old_size: int, new_size: int):
        self.setColumnWidth(logical_index, new_size)
        self.update_geometry()

    def resize_parent_rows(self, logical_index: int, old_size: int, new_size: int):
        if logical_index <= 1:
            self.parent().setRowHeight(logical_index, new_size)
            self.update_geometry()

    def resize_rows(self, logical_index: int, old_size: int, new_size: int):
        if logical_index <= 1:
            self.setRowHeight(logical_index, new_size)
            self.update_geometry()

    def hide_lanes(self):
        for row in range(2, self.model().rowCount()):
            self.hideRow(row)

    def update_geometry(self):
        parent = self.parent()
        assert isinstance(parent, TimeLanesView)

        self.setGeometry(
            parent.viewport().x()
            - parent.verticalHeader().width()
            - parent.frameWidth(),
            parent.horizontalHeader().height(),
            parent.width(),
            parent.rowHeight(0) + parent.rowHeight(1) + parent.frameWidth(),
        )


class TimeLanesView(QTableView):
    def __init__(
        self,
        extension: CondetrolLaneExtensionProtocol,
        device_configurations: dict[DeviceName, DeviceConfiguration],
        parent: Optional[QWidget] = None,
    ):
        """A widget for editing time lanes."""

        super().__init__(parent)
        self._model = TimeLanesModel(extension, self)
        self._device_configurations: dict[DeviceName, DeviceConfiguration] = (
            device_configurations
        )
        self._extension = extension
        self.setModel(self._model)

        self.horizontalHeader().setContextMenuPolicy(
            Qt.ContextMenuPolicy.CustomContextMenu
        )
        self.verticalHeader().setContextMenuPolicy(
            Qt.ContextMenuPolicy.CustomContextMenu
        )
        self.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)
        self.setup_connections()

        # self.setSelectionBehavior(QTableView.SelectionBehavior.SelectItems)
        self.setSelectionMode(QTableView.SelectionMode.ContiguousSelection)

        # We disable the auto-scrolling because it is annoying when the view jumps
        # around on its own while the user is trying to edit the time lanes.
        self.setAutoScroll(False)

        self._steps_table = OverlayStepsView(self._model, self)
        self.viewport().stackUnder(self._steps_table)
        self._steps_table.show()
        self.show()
        self._steps_table.update_geometry()
        self.setVerticalScrollMode(QTableView.ScrollMode.ScrollPerPixel)
        self.verticalHeader().setFixedWidth(200)
        self._steps_table.verticalHeader().setFixedWidth(200)

    def model(self) -> TimeLanesModel:
        model = super().model()
        assert isinstance(model, TimeLanesModel)
        return model

    @property
    def undo_stack(self) -> QUndoStack:
        return self._model.undo_stack

    def moveCursor(self, cursorAction, modifiers):  # noqa: N802, N803
        current = super().moveCursor(cursorAction, modifiers)

        if cursorAction == QTableView.CursorAction.MoveUp:
            if (current.row() > 1) and (
                self.visualRect(current).topLeft().y()
                < self._steps_table.rowHeight(0) + self._steps_table.rowHeight(1)
            ):
                new_value = (
                    self.verticalScrollBar().value()
                    + self.visualRect(current).topLeft().y()
                    - self._steps_table.rowHeight(0)
                    - self._steps_table.rowHeight(1)
                )
                self.verticalScrollBar().setValue(new_value)
            elif current.row() <= 1:
                return current.siblingAtRow(2)
        return current

    def resizeEvent(self, event):  # noqa: N802
        self._steps_table.update_geometry()
        super().resizeEvent(event)

    def setup_connections(self):
        self.horizontalHeader().customContextMenuRequested.connect(
            self.show_steps_context_menu
        )
        self.verticalHeader().customContextMenuRequested.connect(
            self.show_lanes_context_menu
        )
        self.customContextMenuRequested.connect(self.show_cell_context_menu)

        self._model.dataChanged.connect(self.on_data_changed)
        self._model.rowsInserted.connect(self.on_time_lanes_changed)
        self._model.rowsRemoved.connect(self.on_time_lanes_changed)
        self._model.columnsInserted.connect(self.on_time_lanes_changed)
        self._model.columnsRemoved.connect(self.on_time_lanes_changed)
        self._model.modelReset.connect(self.on_time_lanes_changed)

        # only need to update the lane delegates when the rows change
        self._model.rowsInserted.connect(self.update_delegates)
        self._model.rowsRemoved.connect(self.update_delegates)
        self._model.modelReset.connect(self.update_delegates)

    def on_time_lanes_changed(self):
        self.update_spans()

    def on_data_changed(self, top_left: QModelIndex, bottom_right: QModelIndex):
        self.on_time_lanes_changed()

        for row in range(top_left.row(), bottom_right.row() + 1):
            for column in range(top_left.column(), bottom_right.column() + 1):
                index = self._model.index(row, column, QModelIndex())
                span = self._model.span(index)
                if span.width() >= 1 or span.height() >= 1:
                    self.setSpan(row, column, span.height(), span.width())

    def get_time_lanes(self) -> TimeLanes:
        return self._model.get_timelanes()

    def set_time_lanes(self, time_lanes: TimeLanes) -> None:
        self._model.set_timelanes(time_lanes)

    def update_spans(self):
        self.clearSpans()
        for row in range(self._model.rowCount()):
            for column in range(self._model.columnCount()):
                index = self._model.index(row, column, QModelIndex())
                span = self._model.span(index)
                if span.width() >= 1 or span.height() >= 1:
                    self.setSpan(row, column, span.height(), span.width())

    def update_delegates(self):
        for row in range(self._model.rowCount()):
            previous_delegate = self.itemDelegateForRow(row)
            if previous_delegate:
                previous_delegate.deleteLater()
            self.setItemDelegateForRow(row, None)  # type: ignore[reportArgumentType]
        for row in range(2, self._model.rowCount()):
            lane = self._model.get_lane(row - 2)
            name = self._model.get_lane_name(row - 2)
            delegate = self._construct_delegate(lane, name)
            self.setItemDelegateForRow(row, delegate)  # type: ignore[reportArgumentType]
            if delegate:
                delegate.setParent(self)

    def _construct_delegate(
        self, lane: TimeLane, lane_name: str
    ) -> Optional[TimeLaneDelegate]:
        delegate = self._extension.get_lane_delegate(lane, lane_name)
        if delegate is not None:
            if not isinstance(delegate, TimeLaneDelegate):
                raise TypeError(
                    f"Invalid delegate type: {type(delegate)}. "
                    "The delegate must be an instance of TimeLaneDelegate."
                )
            delegate.setParent(self)
            delegate.set_device_configurations(self._device_configurations)
            delegate.set_parameter_names(set())
        return delegate

    def set_read_only(self, read_only: bool) -> None:
        self._model.set_read_only(read_only)

    def is_read_only(self) -> bool:
        return self._model.is_read_only()

    def show_steps_context_menu(self, pos):
        if self.is_read_only():
            return
        menu = QMenu(self.horizontalHeader())

        index = self.horizontalHeader().logicalIndexAt(pos.x())
        if index == -1:
            add_step_action = QAction("Add step")
            menu.addAction(add_step_action)
            add_step_action.triggered.connect(
                lambda: self._model.insertColumn(
                    self._model.columnCount(), QModelIndex()
                )
            )
        elif 0 <= index < self.model().columnCount():
            add_step_before_action = QAction("Insert step before")
            menu.addAction(add_step_before_action)
            add_step_before_action.triggered.connect(
                lambda: self._model.insertColumn(index, QModelIndex())
            )

            add_step_after_action = QAction("Insert step after")
            menu.addAction(add_step_after_action)
            add_step_after_action.triggered.connect(
                lambda: self._model.insertColumn(index + 1, QModelIndex())
            )
            if self.model().columnCount() > 1:
                remove_step_action = QAction("Remove")
                menu.addAction(remove_step_action)
                remove_step_action.triggered.connect(
                    lambda: self._model.removeColumn(index, QModelIndex())
                )
        menu.exec(self.horizontalHeader().mapToGlobal(pos))
        menu.deleteLater()

    def show_lanes_context_menu(self, pos):
        if self.is_read_only():
            return
        menu = QMenu(self.verticalHeader())

        index = self.verticalHeader().logicalIndexAt(pos.y())
        if 2 <= index < self.model().rowCount():
            remove_lane_action = QAction("Remove")
            menu.addAction(remove_lane_action)
            remove_lane_action.triggered.connect(
                lambda: self._model.remove_lane(index - 2)
            )
            for action in self._model.get_lane_header_context_actions(index - 2):
                if isinstance(action, QAction):
                    menu.addAction(action)
                elif isinstance(action, QMenu):
                    menu.addMenu(action)
        else:
            return
        menu.exec(self.verticalHeader().mapToGlobal(pos))
        menu.deleteLater()

    def show_cell_context_menu(self, pos):
        if self.is_read_only():
            return
        index = self.indexAt(pos)
        cell_actions = self._model.get_cell_context_actions(index)
        selection = self.selectionModel().selection()

        with temporary_widget(QMenu(self)) as menu:
            if selection.contains(index):
                merge_action = menu.addAction(f"Expand step {index.column()}")
                merge_action.triggered.connect(
                    lambda: self.expand_step(index.column(), selection)
                )

            for action in cell_actions:
                if isinstance(action, QAction):
                    menu.addAction(action)
                elif isinstance(action, QMenu):
                    menu.addMenu(action)
            menu.exec(self.viewport().mapToGlobal(pos))

    def expand_step(self, step: int, selection):
        indices: set[tuple[int, int]] = set()
        for selection_range in selection:
            top_left = selection_range.topLeft()
            bottom_right = selection_range.bottomRight()
            indices.update(
                itertools.product(
                    range(top_left.row(), bottom_right.row() + 1),
                    range(top_left.column(), bottom_right.column() + 1),
                )
            )

        for row, group in itertools.groupby(sorted(indices), key=lambda x: x[0]):
            group = list(group)
            start = group[0][1]
            stop = group[-1][1]
            self._model.expand_step(step, row - 2, start, stop)

    def add_lane(self, lane_name: str, lane: TimeLane):
        self._model.insert_time_lane(lane_name, lane)

    def visualRect(self, index):  # noqa: N802
        rect = super().visualRect(index)
        if not isinstance(rect, QRect):
            # Not sure this case can actually happen, maybe rect can be None?
            return rect

        # Here we make sure that the rect is within the viewport horizontally.
        # This is useful because when a delegate creates an editor widget, it uses
        # the visualRect method to position the editor.
        # We want the editor to always be fully visible in the viewport, even if the
        # cell is only partially visible.
        viewport_rect = self.viewport().rect()
        if rect.left() < viewport_rect.left():
            rect.setLeft(viewport_rect.left())
        if rect.right() > viewport_rect.right():
            rect.setRight(viewport_rect.right())
        return rect



from typing import Optional, Any, assert_never

from PySide6.QtCore import QObject, QModelIndex, Qt, QPersistentModelIndex
from PySide6.QtGui import QAction
from PySide6.QtWidgets import QMenu

from caqtus.types.expression import Expression
from caqtus.types.timelane import AnalogTimeLane, Ramp
from .._colored_time_lane_model import ColoredTimeLaneModel

_DEFAULT_INDEX = QModelIndex()


class AnalogTimeLaneModel(ColoredTimeLaneModel[AnalogTimeLane]):
    # ruff: noqa: N802
    def __init__(self, name: str, parent: Optional[QObject] = None):
        lane = AnalogTimeLane([Expression("...")])
        super().__init__(name, lane, parent)

    def data(self, index, role: int = Qt.ItemDataRole.DisplayRole):
        if not index.isValid():
            return None
        value = self.lane_value(index.row())
        if role == Qt.ItemDataRole.DisplayRole:
            if isinstance(value, Expression):
                return str(value)
            elif isinstance(value, Ramp):
                return "\u279F"
            else:
                assert_never(value)
        elif role == Qt.ItemDataRole.EditRole:
            if isinstance(value, Expression):
                return str(value)
            return None
        elif role == Qt.ItemDataRole.TextAlignmentRole:
            return Qt.AlignmentFlag.AlignCenter
        else:
            return super().data(index, role)

    def flags(self, index) -> Qt.ItemFlag:
        if not index.isValid():
            return Qt.ItemFlag.NoItemFlags
        flags = super().flags(index)
        if isinstance(self.lane_value(index.row()), Ramp):
            flags &= ~Qt.ItemFlag.ItemIsEditable
        return flags

    def setData(self, index, value: Any, role: int = Qt.ItemDataRole.EditRole):
        if not index.isValid():
            return False
        if role == Qt.ItemDataRole.EditRole:
            if isinstance(value, str):
                new_value = Expression(value)
            elif isinstance(value, Ramp):
                new_value = value
            else:
                raise AssertionError(f"Invalid type for value: {type(value)}")
            return self.set_lane_value(index.row(), new_value)
        return False

    def insertRow(
        self, row, parent: QModelIndex | QPersistentModelIndex = _DEFAULT_INDEX
    ) -> bool:
        return self.insert_lane_value(row, Expression("..."))

    def get_cell_context_actions(self, index: QModelIndex) -> list[QAction | QMenu]:
        if not index.isValid():
            return []

        cell_type_menu = QMenu("Cell type")
        value = self.lane_value(index.row())
        expr_action = cell_type_menu.addAction("expression")
        if isinstance(value, Expression):
            expr_action.setCheckable(True)
            expr_action.setChecked(True)
        else:
            expr_action.triggered.connect(
                lambda: self.setData(index, "...", Qt.ItemDataRole.EditRole)
            )
        ramp_action = cell_type_menu.addAction("ramp")
        if isinstance(value, Ramp):
            ramp_action.setCheckable(True)
            ramp_action.setChecked(True)
        else:
            ramp_action.triggered.connect(
                lambda: self.setData(index, Ramp(), Qt.ItemDataRole.EditRole)
            )

        return super().get_cell_context_actions(index) + [cell_type_menu]
